######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains utilities to scan for Data Transformer's output resources."""
from spine_engine.project_item.project_item_resource import database_resource
from spine_items.utils import database_label
from spinedb_api import append_filter_config
from spinedb_api.filters.tools import store_filter


def scan_for_resources(provider, dt_specification, db_resources, filter_config_path):
    """
    Returns a list of resources, i.e., filtered dbs produced by the Data Transformer.

    Args:
        provider (ProjectItem or ExecutableItem): resource provider item
        dt_specification (DataTransformerSpecification): spec
        db_resources (list of ProjectItemResource): db resources available to the DT
        filter_config_path (str): path to filter config file

    Returns:
        list: a list of output resources
    """
    label = database_label(provider.name)
    if dt_specification is None or dt_specification.settings is None:
        return [database_resource(provider.name, r.url, label=label) for r in db_resources]
    if dt_specification.settings.use_shorthand():
        config = dt_specification.settings.filter_config()
        return [
            database_resource(provider.name, append_filter_config(r.url, config), label=label) for r in db_resources
        ]
    with open(filter_config_path, "w") as filter_config_file:
        store_filter(dt_specification.settings.filter_config(), filter_config_file)
    return [
        database_resource(provider.name, append_filter_config(r.url, filter_config_path), label=label)
        for r in db_resources
    ]



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Data Store's executable item as well as support utilities."""
from pathlib import Path
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.utils.serialization import deserialize_path
from spinedb_api import DatabaseMapping
from spinedb_api.exception import SpineDBAPIError
from ..utils import check_database_url, convert_to_sqlalchemy_url
from .item_info import ItemInfo
from .output_resources import scan_for_resources


class ExecutableItem(ExecutableItemBase):
    def __init__(self, name, url, project_dir, logger):
        """
        Args:
            name (str): item's name
            url (URL): database's URL
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        self._url = url
        self._validated = False

    @staticmethod
    def item_type():
        """Returns the data store executable's type identifier string."""
        return ItemInfo.item_type()

    @staticmethod
    def is_filter_terminus():
        """See base class."""
        return True

    def ready_to_execute(self, settings):
        """See base class."""
        if not super().ready_to_execute(settings) or not self._check_sqlite_file_exists():
            return False
        error = check_database_url(self._url)
        return error is None

    def _get_url(self):
        if not self._check_sqlite_file_exists():
            self._logger.msg_error.emit("SQLite file does not exist.")
            return None
        if not self._validated:
            prompt_data = DatabaseMapping.get_upgrade_db_prompt_data(self._url, create=True)
            if prompt_data is not None:
                kwargs = self._logger.prompt.emit(prompt_data)
                if kwargs is None:
                    return None
            else:
                kwargs = {}
            try:
                DatabaseMapping.create_engine(self._url, create=True, **kwargs)
                return self._url
            except SpineDBAPIError as err:
                self._logger.msg_error.emit(str(err))
                return None
            finally:
                self._validated = True
        return self._url

    def _check_sqlite_file_exists(self):
        """Checks that the database file exists for SQLite databases.

        Returns:
            bool: True if SQLite file exists or database has different dialect, False otherwise
        """
        if self._url.drivername.startswith("sqlite"):
            database_path = Path(self._url.database)
            if not database_path.exists():
                return False
            if database_path.is_dir():
                return False
        return True

    def _output_resources_backward(self):
        """See base class."""
        return self._output_resources_forward()

    def _output_resources_forward(self):
        """See base class."""
        return scan_for_resources(self, self._get_url())

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        if item_dict["url"]["dialect"] == "sqlite":
            item_dict["url"]["database"] = deserialize_path(item_dict["url"]["database"], project_dir)
        url = convert_to_sqlalchemy_url(item_dict["url"], name, logger)
        return cls(name, url, project_dir, logger)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains undo and redo commands for Importer specification editor."""
from enum import IntEnum, auto, unique
import pickle
from PySide6.QtCore import QModelIndex, Qt
from PySide6.QtGui import QUndoCommand
from spine_items.importer.mvcmodels.mappings_model_roles import Role
from spinedb_api import import_mapping_from_dict
from spinedb_api.import_mapping.import_mapping import from_dict as mapping_from_dict


@unique
class _Id(IntEnum):
    SET_OPTION = auto()


class PasteMappings(QUndoCommand):
    """Command to paste copied mappings"""

    def __init__(self, table_row, list_row, model, mapping_list_data, text="paste mappings"):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            mapping_list_data (MappingListMimeData): mappings to paste
            text (str): command text
        """
        super().__init__(text)
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._mapping_list_data = mapping_list_data

    def redo(self):
        """Pastes the copied mappings."""
        table_index = self._model.index(self._table_row, 0)
        for i, mapping_data in enumerate(self._mapping_list_data.mapping_list):
            root_mapping = mapping_from_dict(mapping_data.mapping_dict)
            row = self._list_row + i
            self._model.insertRow(row, table_index)
            self._model.rename_mapping(self._table_row, row, mapping_data.name)
            self._model.set_root_mapping(self._table_row, row, root_mapping)

    def undo(self):
        """Removes pasted mappings."""
        table_index = self._model.index(self._table_row, 0)
        self._model.removeRows(self._list_row, len(self._mapping_list_data.mapping_list), table_index)


class MoveTableInList(QUndoCommand):
    """Command to move source tables in the list."""

    def __init__(self, model, source_row, destination_row):
        """
        Args:
            model (MappingsModel): model
            source_row (int): source table list row index to move
            destination_row (int): target source table list row index
        """
        super().__init__("move table")
        self._model = model
        self._source_row = source_row
        self._destination_row = destination_row

    def redo(self):
        self._model.moveRow(QModelIndex(), self._source_row, QModelIndex(), self._destination_row)

    def undo(self):
        source = self._destination_row - (1 if self._source_row < self._destination_row else 0)
        destination = self._source_row + (1 if self._source_row > self._destination_row else 0)
        self._model.moveRow(QModelIndex(), source, QModelIndex(), destination)


class MoveMappingInList(QUndoCommand):
    """Command to move mapping in mapping list."""

    def __init__(self, table_row, model, source_row, destination_row):
        """
        Args:
            table_row (int): source table row index
            model (MappingsModel): model
            source_row (int): mapping list row index to move
            destination_row (int): target mapping list row index
        """
        super().__init__("move mapping")
        self._table_row = table_row
        self._model = model
        self._source_row = source_row
        self._destination_row = destination_row

    def redo(self):
        parent = self._model.index(self._table_row, 0)
        self._model.moveRow(parent, self._source_row, parent, self._destination_row)

    def undo(self):
        parent = self._model.index(self._table_row, 0)
        source = self._destination_row - (1 if self._source_row < self._destination_row else 0)
        destination = self._source_row + (1 if self._source_row > self._destination_row else 0)
        self._model.moveRow(parent, source, parent, destination)


class PasteOptions(QUndoCommand):
    """Command to paste copied mapping options."""

    def __init__(self, import_sources, source_table_name, pickled_options, previous_options):
        """
        Args:
            import_sources (ImportSources): import editor
            source_table_name (str): name of the target source table
            pickled_options (bytes): pickled options
            previous_options (bytes): pickled previous options
        """
        super().__init__("paste options")
        self._import_sources = import_sources
        self._source_table_name = source_table_name
        self._pickled_options = pickled_options
        self._previous_options = previous_options

    def redo(self):
        """Pastes the options."""
        self._import_sources.paste_options(self._source_table_name, pickle.loads(self._pickled_options))

    def undo(self):
        """Restores the options to their previous values."""
        self._import_sources.paste_options(self._source_table_name, pickle.loads(self._previous_options))


class SetTableChecked(QUndoCommand):
    """Command to change a source table's checked state."""

    def __init__(self, table_name, model, checked, *rows):
        """
        Args:
            table_name (str, optional): source table name
            model (MappingsModel): model
            checked (bool): new checked state
            *rows: table row on the list
        """
        if table_name is not None:
            text = ("select" if checked else "deselect") + f" '{table_name}'"
        else:
            text = ("select" if checked else "deselect") + " all source tables'"
        super().__init__(text)
        self._model = model
        self._rows = rows
        self._checked = checked

    def redo(self):
        """Changes the checked state."""
        self._model.set_table_checked(self._checked, *self._rows)

    def undo(self):
        """Restores the previous checked state."""
        self._model.set_table_checked(not self._checked, *self._rows)


class UpdateTableItem(QUndoCommand):
    """Command to update table name and state in file-less mode."""

    def __init__(self, model, row, old_data, new_data, is_empty_row):
        """
        Args:
            model (MappingsModel): model
            row (int): table row on the list
            old_data (dict): old item data
            new_data (dict): updated item data
            is_empty_row (bool): True if the table doesn't contain any mappings
        """
        text = "rename table" if not is_empty_row else "add table"
        super().__init__(text)
        self._model = model
        self._row = row
        self._old_data = old_data
        self._new_data = new_data
        self._is_empty_row = is_empty_row

    def redo(self):
        self._model.update_table_item(self._row, self._new_data, add_empty_row=self._is_empty_row)

    def undo(self):
        self._model.update_table_item(self._row, self._old_data, remove_empty_row=self._is_empty_row)


class SetMappingPositionType(QUndoCommand):
    """Sets the type of component mapping."""

    def __init__(self, table_row, list_row, row, model, position_type, previous_type, previous_position):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            model (MappingsModel): model
            position_type (str): name of the new type
            previous_type (str): name of the original position type
            previous_position (str or int): original position
        """
        super().__init__("change source mapping")
        self._table_row = table_row
        self._list_row = list_row
        self._row = row
        self._model = model
        self._new_type = position_type
        self._previous_type = previous_type
        self._previous_position = previous_position

    def redo(self):
        """Changes a component mapping's type."""
        self._model.set_mapping_position_type(self._table_row, self._list_row, self._row, self._new_type)

    def undo(self):
        """Restores component mapping's original type."""
        self._model.set_mapping_position_type(self._table_row, self._list_row, self._row, self._previous_type)
        self._model.set_mapping_position(
            self._table_row, self._list_row, self._row, self._previous_type, self._previous_position
        )


class SetMappingPosition(QUndoCommand):
    """Sets the reference for a component mapping."""

    def __init__(self, table_row, list_row, row, model, new_type, new_position, previous_type, previous_position):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            model (MappingsModel): model
            new_type (str): new position type
            new_position (str or int): new value for the position
            previous_type (str): original position type
            previous_position (str or int): position's original value
        """
        super().__init__("change source mapping reference")
        self._table_row = table_row
        self._list_row = list_row
        self._row = row
        self._model = model
        self._new_type = new_type
        self._new_position = new_position
        self._previous_type = previous_type
        self._previous_position = previous_position

    def redo(self):
        """Sets the reference's value."""
        self._model.set_mapping_position(self._table_row, self._list_row, self._row, self._new_type, self._new_position)

    def undo(self):
        """Restores the reference's value and, if necessary, mapping type to their original values."""
        if self._previous_type == "None":
            self._model.set_mapping_position_type(self._table_row, self._list_row, self._row, "None")
        else:
            self._model.set_mapping_position(
                self._table_row, self._list_row, self._row, self._previous_type, self._previous_position
            )


class SetFilterRe(QUndoCommand):
    """Sets mapping component's filter regular expression."""

    def __init__(self, table_row, list_row, row, model, new_re, previous_re):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            row (int): mapping component row index
            model (MappingsModel): model
            new_re (str): new filter regexp
            previous_re (str): previous filter regexp
        """
        super().__init__("change source mapping reference")
        self._table_row = table_row
        self._list_row = list_row
        self._row = row
        self._model = model
        self._re = new_re
        self._previous_re = previous_re

    def redo(self):
        self._model.set_filter_re(self._table_row, self._list_row, self._row, self._re)

    def undo(self):
        self._model.set_filter_re(self._table_row, self._list_row, self._row, self._previous_re)


class SetReaderOption(QUndoCommand):
    """Command to set a :class:`ConnectorManager` option."""

    def __init__(self, source_table, option_key, options_widget, value, previous_value):
        """
        Args:
            source_table (str): source table name
            option_key (str): option's key
            options_widget (OptionsWidget): connector options widget
            value (str or int or bool): option's new value
            previous_value (str or int or bool): option's previous value
        """
        text = f"change {option_key}"
        super().__init__(text)
        self._source_table = source_table
        self._option_key = option_key
        self._options_widget = options_widget
        self._value = value
        self._previous_value = previous_value

    def id(self):
        """
        This command's id.

        Returns:
            int: id
        """
        return _Id.SET_OPTION

    def mergeWith(self, command):
        """
        Merges command with another :class:`SetConnectorOption`.

        Args:
            command (QUndoCommand): a command to merge with

        Returns:
            bool: True if merge was successful, False otherwise
        """
        if not isinstance(command, SetReaderOption):
            return False
        return command._option_key == self._option_key and command._value == self._value

    def redo(self):
        """Changes the connector's option."""
        self._options_widget.set_option_without_undo(self._source_table, self._option_key, self._value)

    def undo(self):
        """Restores the option back to its original value."""
        self._options_widget.set_option_without_undo(self._source_table, self._option_key, self._previous_value)


class CreateMapping(QUndoCommand):
    """Creates a new mapping."""

    def __init__(self, table_row, model, list_row):
        """
        Args:
            table_row (int): source table row index
            model (MappingsModel): model
            list_row (int): mapping list row where the new mapping should be created
        """
        super().__init__("new mapping")
        self._table_row = table_row
        self._model = model
        self._list_row = list_row

    def redo(self):
        """Creates a new mapping at the given row in mappings list."""
        table_index = self._model.index(self._table_row, 0)
        self._model.insertRow(self._list_row, table_index)

    def undo(self):
        """Deletes the created mapping."""
        table_index = self._model.index(self._table_row, 0)
        self._model.removeRow(self._list_row, table_index)


class RenameMapping(QUndoCommand):
    """Renames a mapping."""

    def __init__(self, table_row, row, model, name, previous_name):
        """
        Args:
            table_row (int): source table row
            row (int): mapping list row
            model (MappingsModel): model
            name (str): new name
            previous_name (str): previous name
        """
        super().__init__("new mapping")
        self._table_row = table_row
        self._row = row
        self._model = model
        self._name = name
        self._previous_name = previous_name

    def redo(self):
        self._model.rename_mapping(self._table_row, self._row, self._name)

    def undo(self):
        self._model.rename_mapping(self._table_row, self._row, self._previous_name)


class DuplicateMapping(QUndoCommand):
    """Duplicates an existing mapping."""

    def __init__(self, table_row, model, list_row):
        """
        Args:
            table_row (int): source table row index
            model (MappingsModel): model
            list_row (int): mapping list row where the new mapping should be created
        """
        super().__init__("duplicate mapping")
        self._table_row = table_row
        self._model = model
        self._list_row = list_row

    def redo(self):
        """Duplicates a mapping below the original one in mapping list."""
        table_index = self._model.index(self._table_row, 0)
        list_index = self._model.index(self._list_row, 0, table_index)
        root_mapping = list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        mapping_dict = [m.to_dict() for m in root_mapping.flatten()]
        self._model.insertRow(self._list_row + 1, table_index)
        self._model.set_root_mapping(self._table_row, self._list_row + 1, import_mapping_from_dict(mapping_dict))

    def undo(self):
        """Deletes the duplicated mapping."""
        table_index = self._model.index(self._table_row, 0)
        self._model.removeRow(self._list_row + 1, table_index)


class DeleteMapping(QUndoCommand):
    """Command to delete a mapping."""

    def __init__(self, table_row, model, list_row):
        """
        Args:
            table_row (int): source table row index
            model (MappingsModel): model
            list_row (int): mapping list row which should be deleted
        """
        super().__init__("delete mapping")
        self._table_row = table_row
        self._model = model
        self._list_row = list_row
        table_index = self._model.index(table_row, 0)
        list_index = self._model.index(list_row, 0, table_index)
        name = list_index.data()
        root_mapping = list_index.data(Role.FLATTENED_MAPPINGS).root_mapping
        self._deleted_mapping = name, [m.to_dict() for m in root_mapping.flatten()]

    def redo(self):
        """Deletes the mapping."""
        self._model.removeRow(self._list_row, self._model.index(self._table_row, 0))

    def undo(self):
        """Restores the deleted mapping."""
        table_index = self._model.index(self._table_row, 0)
        self._model.insertRow(self._list_row, table_index)
        name, mapping_dict = self._deleted_mapping
        mapping_list_index = self._model.index(self._list_row, 0, table_index)
        self._model.setData(mapping_list_index, name)
        self._model.set_root_mapping(self._table_row, self._list_row, import_mapping_from_dict(mapping_dict))


class SetItemMappingType(QUndoCommand):
    """Command to change item mapping's type."""

    def __init__(self, table_row, list_row, model, new_type, previous_mapping):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            new_type (str): name of the new mapping type
            previous_mapping (ImportMapping): the previous root mapping
        """
        super().__init__("mapping type change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._new_type = new_type
        self._previous_mapping_dict = [m.to_dict() for m in previous_mapping.flatten()]

    def redo(self):
        """Sets the mapping type to its new value."""
        self._model.set_mappings_type(self._table_row, self._list_row, self._new_type)

    def undo(self):
        """Resets the mapping."""
        self._model.set_root_mapping(
            self._table_row, self._list_row, import_mapping_from_dict(self._previous_mapping_dict)
        )


class SetImportEntitiesFlag(QUndoCommand):
    """Command to set item mapping's import objects flag."""

    def __init__(self, table_row, list_row, model, import_entities):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            import_entities (bool): new flag value
        """
        super().__init__(("check" if import_entities else "uncheck") + " import entities")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._import_entities = import_entities

    def redo(self):
        """Changes the import objects flag."""
        self._model.set_import_entities(self._table_row, self._list_row, self._import_entities)

    def undo(self):
        """Restores the import objects flag."""
        self._model.set_import_entities(self._table_row, self._list_row, not self._import_entities)


class SetImportEntityAlternatives(QUndoCommand):
    """Command to add/remove entity alternative imports to/from mappings."""

    def __init__(self, table_row, list_row, model, import_entity_alternatives, previous_mapping):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            import_entity_alternatives (bool): new flag value
            previous_mapping (ImportMapping): previous mapping root
        """
        super().__init__(("check" if import_entity_alternatives else "uncheck") + " import entities")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._import_entity_alternatives = import_entity_alternatives
        self._previous_mapping_dict = [m.to_dict() for m in previous_mapping.flatten()]

    def redo(self):
        """Changes the ."""
        self._model.set_import_entity_alternatives(self._table_row, self._list_row, self._import_entity_alternatives)

    def undo(self):
        """Restores the mappings to previous state."""
        self._model.set_root_mapping(
            self._table_row, self._list_row, import_mapping_from_dict(self._previous_mapping_dict)
        )


class SetParameterType(QUndoCommand):
    """Command to change the parameter type of item mapping."""

    def __init__(self, table_row, list_row, model, new_type, previous_mapping):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            new_type (str): name of the new parameter type
            previous_mapping (ImportMapping): previous mapping root
        """
        super().__init__("parameter type change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._new_type = new_type
        self._previous_mapping_dict = [m.to_dict() for m in previous_mapping.flatten()]

    def redo(self):
        """Changes a parameter's type."""
        self._model.set_parameter_type(self._table_row, self._list_row, self._new_type)

    def undo(self):
        """Restores a parameter to its previous type"""
        self._model.set_root_mapping(
            self._table_row, self._list_row, import_mapping_from_dict(self._previous_mapping_dict)
        )


class SetValueType(QUndoCommand):
    """Command to change the value type of item mapping."""

    def __init__(self, table_row, list_row, model, new_type, old_type):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            new_type (str): name of the new value type
            old_type (str): name of the old value type
        """
        super().__init__("value type change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._new_type = new_type
        self._old_type = old_type

    def redo(self):
        """Changes a parameter's value type."""
        self._model.set_value_type(self._table_row, self._list_row, self._new_type)

    def undo(self):
        """Restores a parameter to its previous value type"""
        self._model.set_value_type(self._table_row, self._list_row, self._old_type)


class SetSkipColumns(QUndoCommand):
    """Command to change item mapping's skip columns option."""

    def __init__(self, table_row, list_row, model, skip_cols, previous_skip_cols):
        """
        Args:
            table_row (int): source table row index
            list_row (int): mapping list row index
            model (MappingsModel): model
            skip_cols (list): new skip columns
            previous_skip_cols (list): previous skip columns
        """
        super().__init__("mapping ignore columns change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._skip_cols = skip_cols
        self._previous_skip_cols = previous_skip_cols

    def redo(self):
        """Changes item mapping's skip columns to a new value."""
        self._model.set_skip_columns(self._table_row, self._list_row, self._skip_cols)

    def undo(self):
        """Restores item mapping's skip columns to its previous value."""
        self._model.set_skip_columns(self._table_row, self._list_row, self._previous_skip_cols)


class SetReadStartRow(QUndoCommand):
    """Command to change item mapping's read start row option."""

    def __init__(self, table_row, list_row, model, start_row, previous_start_row):
        """
        Args:
            table_row (int): table row index
            list_row (int): list row index
            model (MappingsModel): model
            start_row (int): new read start row
            previous_start_row (int): previous read start row value
        """
        super().__init__("mapping read start row change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._start_row = start_row
        self._previous_start_row = previous_start_row

    def redo(self):
        """Changes item mapping's read start row to a new value."""
        self._model.set_read_start_row(self._table_row, self._list_row, self._start_row)

    def undo(self):
        """Restores item mapping's read start row to its previous value."""
        self._model.set_read_start_row(self._table_row, self._list_row, self._previous_start_row)


class SetItemMappingDimensionCount(QUndoCommand):
    """Command to change item mapping's dimension option."""

    def __init__(self, table_row, list_row, model, dimension_count, previous_mapping_root):
        """
        Args:
            table_row (int): table row index
            list_row (int): list row index
            model (MappingsModel): model
            dimension_count (int): new dimension count
            previous_mapping_root (ImportMapping): previous mapping root
        """
        super().__init__("mapping dimension count change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._dimension_count = dimension_count
        self._previous_mapping_dict = [m.to_dict() for m in previous_mapping_root.flatten()]

    def redo(self):
        """Changes the item mapping's dimension to the new value."""
        self._model.set_relationship_dimension_count(self._table_row, self._list_row, self._dimension_count)

    def undo(self):
        """Changes the item mapping's dimension to its previous value."""
        self._model.set_root_mapping(
            self._table_row, self._list_row, import_mapping_from_dict(self._previous_mapping_dict)
        )


class SetTimeSeriesRepeatFlag(QUndoCommand):
    """Command to change the repeat flag for time series."""

    def __init__(self, table_row, list_row, model, repeat):
        """
        Args:
            table_row (int): table row index
            list_row (int): list row index
            model (MappingsModel): model
            repeat (bool): new repeat flag value
        """
        super().__init__("change time series repeat flag")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._repeat = repeat

    def redo(self):
        """Sets the repeat flag to given value."""
        self._model.set_time_series_repeat_flag(self._table_row, self._list_row, self._repeat)

    def undo(self):
        """Restores the repeat flag to its previous value."""
        self._model.set_time_series_repeat_flag(self._table_row, self._list_row, not self._repeat)


class SetMapDimensionCount(QUndoCommand):
    """Command to change the dimension_count of a Map parameter value type."""

    def __init__(self, table_row, list_row, model, dimension_count, previous_mapping_root):
        """
        Args:
            table_row (int): table row index
            list_row (int): list row index
            model (MappingsModel): model
            dimension_count (int): new dimension_count
            previous_mapping_root (ImportMapping): previous mapping root
        """
        super().__init__("map dimension count change")
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._dimension_count = dimension_count
        self._previous_mapping_root = [m.to_dict() for m in previous_mapping_root.flatten()]

    def redo(self):
        """Sets the Map dimension count to the new value."""
        self._model.set_map_dimension_count(self._table_row, self._list_row, self._dimension_count)

    def undo(self):
        """Restores the previous Map dimension count value."""
        self._model.set_root_mapping(
            self._table_row, self._list_row, import_mapping_from_dict(self._previous_mapping_root)
        )


class SetMapCompressFlag(QUndoCommand):
    """Command to change the Map compress flag."""

    def __init__(self, table_row, list_row, model, compress):
        """
        Args:
            table_row (int): table row index
            list_row (int): list row index
            model (MappingsModel): model
            compress (bool): compress flag value
        """
        text = ("enable" if compress else "disable") + " Map compression"
        super().__init__(text)
        self._table_row = table_row
        self._list_row = list_row
        self._model = model
        self._compress = compress

    def redo(self):
        """Sets the compress flag."""
        self._model.set_map_compress(self._table_row, self._list_row, self._compress)

    def undo(self):
        """Resets the compress flag to previous value."""
        self._model.set_map_compress(self._table_row, self._list_row, not self._compress)


class SetColumnOrRowType(QUndoCommand):
    """Command to change the type of columns or rows."""

    def __init__(self, source_table_name, header_widget, sections, new_type, previous_type):
        """
        Args:
            source_table_name (str): name of the source table
            header_widget (HeaderWithButton): widget of origin
            sections (Iterable of int): row or column indexes
            new_type (ConvertSpec): conversion specification for the rows/columns
            previous_type (ConvertSpec): previous conversion specification for the rows/columns
        """
        text = ("row" if header_widget.orientation() == Qt.Orientation.Vertical else "column") + " type change"
        super().__init__(text)
        self._source_table_name = source_table_name
        self._header_widget = header_widget
        self._sections = sections
        self._new_type = new_type
        self._previous_type = previous_type

    def redo(self):
        """Sets column/row type."""
        self._header_widget.set_data_types(self._source_table_name, self._sections, self._new_type)

    def undo(self):
        """Restores column/row type to its previous value."""
        self._header_widget.set_data_types(self._source_table_name, self._sections, self._previous_type)


class SetColumnDefaultType(QUndoCommand):
    """Command to change the default type of columns."""

    def __init__(self, import_sources, source_table_name, new_type, previous_type):
        """
        Args:
            import_sources (ImportSources): import sources manager
            source_table_name (str): name of the source table
            new_type (str): new column type
            previous_type (str): previous column type
        """
        super().__init__("surplus column type")
        self._import_sources = import_sources
        self._table_name = source_table_name
        self._new_type = new_type
        self._previous_type = previous_type

    def redo(self):
        self._import_sources.change_default_column_type(self._table_name, self._new_type)

    def undo(self):
        self._import_sources.change_default_column_type(self._table_name, self._previous_type)


class RestoreMappingsFromDict(QUndoCommand):
    """Restores mappings from a dict."""

    def __init__(self, import_sources, model, mapping_dict):
        """
        Args:
            import_sources (ImportSources): import sources manager
            model (MappingsModel): model
            mapping_dict (dict): mappings to restore
        """
        super().__init__("import mappings")
        self._import_sources = import_sources
        self._model = model
        self._mapping_dict = mapping_dict
        self._previous_mapping_dict = self._import_sources.store_connectors()
        self._previous_mapping_dict.update(self._model.store())

    def redo(self):
        """Restores the mappings."""
        self._model.restore(self._mapping_dict)
        self._import_sources.restore_connectors(self._mapping_dict)

    def undo(self):
        """Reverts back to previous mappings."""
        self._model.restore(self._previous_mapping_dict)
        self._import_sources.restore_connectors(self._mapping_dict)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Classes for custom context menus and pop-up menus."""
from spinetoolbox.widgets.custom_menus import CustomContextMenu


class DcRefContextMenu(CustomContextMenu):
    """Context menu class for references view in Data Connection properties."""

    def __init__(self, parent, position, index, dc):
        """
        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            index (QModelIndex): Index of item that requested the context-menu
            dc (DataConnection): Data connection item
        """
        super().__init__(parent, position)
        self.add_action("Open...", enabled=dc.current_is_file_ref)
        self.add_action("Open containing directory...", enabled=dc.current_is_file_ref)
        self.addSeparator()
        self.add_action("Add file reference(s)...")
        self.add_action("Add database reference...")
        self.add_action("Remove reference(s)", enabled=dc.any_refs_selected)
        self.add_action("Copy file reference(s) to project", enabled=dc.file_refs_selected)
        self.addSeparator()
        self.add_action("Refresh reference(s)", enabled=dc.any_refs_selected)


class DcDataContextMenu(CustomContextMenu):
    """Context menu class for data view in Data Connection properties."""

    def __init__(self, parent, position, index, dc):
        """
        Args:
            parent (QWidget): Parent for menu widget (ToolboxUI)
            position (QPoint): Position on screen
            index (QModelIndex): Index of item that requested the context-menu
            dc (DataConnection): Data connection item
        """
        super().__init__(parent, position)
        self.add_action("Open...", enabled=index.isValid())
        self.add_action("New file...")
        self.add_action("Remove file(s)", enabled=dc.any_data_selected)
        self.addSeparator()
        self.add_action("Open directory...")



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Utility functions for the Tool project item. """
from datetime import datetime
import glob
import json
import os.path
from pathlib import Path
import re
from spine_engine.utils.helpers import custom_find_kernel_specs, resolve_julia_executable, resolve_julia_project


def get_julia_path_and_project(exec_settings, settings):
    """Returns path to Julia and --project=path/to/project in a list based on Tool specs execution settings.

    Args:
        exec_settings (dict): Execution settings
        settings (AppSettings): application settings

    Returns:
        list of str: e.g. ["path/to/julia", "--project=path/to/project/"] or None if kernel does not exist.
    """
    use_jupyter_console = exec_settings["use_jupyter_console"]
    if use_jupyter_console:
        kernel_name = exec_settings["kernel_spec_name"]
        resource_dir = custom_find_kernel_specs().get(kernel_name)
        if resource_dir is None:
            return None
        filepath = os.path.join(resource_dir, "kernel.json")
        with open(filepath, "r") as fh:
            try:
                kernel_spec = json.load(fh)
            except json.decoder.JSONDecodeError:
                return None
        julia = kernel_spec["argv"].pop(0)
        project_arg = next((arg for arg in kernel_spec["argv"] if arg.startswith("--project=")), None)
        project = "" if project_arg is None else project_arg.split("--project=")[1]
        retval = [julia]
        if project:
            retval.append(f"--project={project}")
        return retval
    julia = exec_settings["executable"]
    if not julia:
        julia = resolve_julia_executable(settings)
    project = exec_settings["project"]
    if not project:
        project = resolve_julia_project(settings)
    retval = [julia]
    if project:
        retval.append(f"--project={project}")
    return retval


def flatten_file_path_duplicates(file_paths, logger, log_duplicates=False):
    """Flattens the extra duplicate dimension in file_paths."""
    flattened = {}
    for required_file, paths in file_paths.items():
        if paths is not None:
            pick = paths[0]
            if len(paths) > 1 and log_duplicates:
                logger.msg_warning.emit(f"Multiple input files satisfy {required_file}; using {pick}")
            flattened[required_file] = pick
        else:
            flattened[required_file] = None
    return flattened


def file_paths_from_resources(resources):
    """
    Returns file paths from given resources.

    Args:
        resources (list): resources available

    Returns:
        a list of file paths, possibly including patterns
    """
    file_paths = []
    glob_chars = ("*", "?", "[")
    for resource in resources:
        if resource.hasfilepath:
            path = resource.path
            if any(char in path for char in glob_chars):
                file_paths += glob.glob(path)
            else:
                file_paths.append(path)
        elif resource.type_ == "file":
            file_paths.append(resource.label)
    return file_paths


def find_file(filename, resources, one_file=None):
    """
    Returns all occurrences of full paths to given file name in resources available.

    Args:
        filename (str): Searched file name (no path)
        resources (list): list of resources available from upstream items
        one_file (bool): If True, a list containing only the first found matching file is returned

    Returns:
        list: Full paths to file if found, None if not found
    """
    filename = os.path.normcase(filename)
    found_file_paths = []
    for file_path in file_paths_from_resources(resources):
        _, file_candidate = os.path.split(file_path)
        if os.path.normcase(file_candidate) == filename:
            found_file_paths.append(file_path)
            if one_file:
                break
    return found_file_paths if found_file_paths else None


def find_last_output_files(output_files, output_dir):
    """
    Returns latest output files.

    Args:
        output_files (list): output file patterns from tool specification
        output_dir (str): path to the execution output directory

    Returns:
        dict: a mapping from a file name pattern to the path of the most recent files in the results archive.
    """
    if not os.path.exists(output_dir):
        return {}
    recent_output_files = {}
    file_patterns = list(output_files)
    result = _find_last_output_dir(output_dir)
    if result is None:
        return {}
    full_archive_path = result[1]
    for pattern in list(file_patterns):
        full_path_pattern = os.path.join(full_archive_path, pattern)
        files_found = False
        for path in glob.glob(full_path_pattern):
            if os.path.exists(path):
                files_found = True
                file_list = recent_output_files.setdefault(pattern, [])
                file_list.append(os.path.normpath(path))
        if files_found:
            file_patterns.remove(pattern)
        if not file_patterns:
            return recent_output_files
    return recent_output_files


def _find_last_output_dir(output_dir, depth=0):
    """Searches for the latest output archive directory recursively.

    Args:
        output_dir (str):  path to the execution output directory
        depth (int): current recursion depth

    Returns:
        tuple: creation datetime and directory path
    """
    latest = None
    result_directory_pattern = re.compile(r"^\d\d\d\d-\d\d-\d\dT\d\d.\d\d.\d\d")
    for path in Path(output_dir).iterdir():
        if not path.is_dir() or path.name == "failed":
            continue
        if result_directory_pattern.match(path.name) is not None:
            time_stamp = datetime.fromisoformat(path.name.replace(".", ":"))
            if latest is None:
                latest = (time_stamp, path)
            elif latest[0] < time_stamp:
                latest = (time_stamp, path)
        elif depth < 1:
            subdir_latest = _find_last_output_dir(str(path), depth + 1)
            if latest is None:
                latest = subdir_latest
            elif subdir_latest is not None and latest[0] < subdir_latest[0]:
                latest = subdir_latest
    return latest


def is_pattern(file_name):
    """Returns True if file_name is actually a file pattern."""
    return "*" in file_name or "?" in file_name


def make_dir_if_necessary(d, directory):
    """Creates a directory if given dictionary contains any items.

    Args:
        d (dict): Dictionary to check
        directory (str): Absolute path to directory that shall be created if necessary

    Returns:
        bool: True if directory was created successfully or dictionary is empty,
        False if creating the directory failed.
    """
    if len(d.items()) > 0:
        try:
            os.makedirs(directory, exist_ok=True)
        except OSError:
            return False
    return True



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains :class:`PreviewUpdater`."""
from copy import deepcopy
from dataclasses import dataclass
from multiprocessing import Process, Queue
import queue
from time import monotonic
from PySide6.QtCore import QItemSelectionModel, QModelIndex, Qt, QTimer, Slot
from PySide6.QtWidgets import QFileDialog
from spinedb_api import DatabaseMapping, SpineDBAPIError, SpineDBVersionError
from spinedb_api.export_mapping.export_mapping import ExportMapping
from spinedb_api.export_mapping.group_functions import GroupFunction
from spinedb_api.spine_io.exporters.writer import write
from ..mvcmodels.full_url_list_model import FullUrlListModel
from ..mvcmodels.mappings_table_model import MappingsTableModel
from ..mvcmodels.preview_table_model import PreviewTableModel
from ..mvcmodels.preview_tree_model import PreviewTreeModel
from ..preview_table_writer import TableWriter


class PreviewUpdater:
    def __init__(
        self, window, ui, url_model, mappings_table_model, mappings_proxy_model, mapping_editor_table_model, project_dir
    ):
        """
        Args:
            window (SpecificationEditorWindow): specification editor's window
            ui (Ui_Form): specification editor's UI
            url_model (FullUrlListModel, optional): URL model
            mappings_table_model (MappingsTableModel): mappings table model
            mappings_proxy_model (QAbstractProxyModel): proxy model for mappings table
            mapping_editor_table_model (MappingEditorTableModel): mapping table model
            project_dir (str): path to initial directory from which to load databases
        """
        self._current_url = None
        self._window = window
        self._ui = ui
        self._project_dir = project_dir
        if url_model is None:
            url_model = FullUrlListModel(self._ui.database_url_combo_box)
        self._url_model = url_model
        self._url_model.rowsInserted.connect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.connect(self._enable_controls)
        self._url_model.destroyed.connect(self._forget_url_model)
        self._ui.database_url_combo_box.setModel(self._url_model)
        self._mappings_table_model = mappings_table_model
        self._set_expect_removals_and_inserts(False)
        self._mappings_table_model.write_order_about_to_change.connect(
            lambda: self._set_expect_removals_and_inserts(True)
        )
        self._mappings_table_model.write_order_changed.connect(lambda: self._set_expect_removals_and_inserts(False))
        self._mappings_table_model.dataChanged.connect(self._rename_mapping)
        self._mappings_table_model.dataChanged.connect(self._enable_mapping)
        self._mappings_table_model.dataChanged.connect(self._update_changed_tables)
        self._mappings_proxy_model = mappings_proxy_model
        self._mapping_editor_table_model = mapping_editor_table_model
        self._mapping_editor_table_model.dataChanged.connect(lambda *_: self._update_current_mappings_tables())
        self._expect_mapping_editor_resets(False)
        self._window.current_mapping_about_to_change.connect(lambda: self._expect_mapping_editor_resets(True))
        self._window.current_mapping_changed.connect(lambda: self._expect_mapping_editor_resets(False))
        self._preview_tree_model = PreviewTreeModel()
        self._preview_tree_model.dataChanged.connect(self._update_table)
        self._preview_tree_model.dataChanged.connect(self._expand_tree_after_table_change)
        self._preview_tree_model.rowsInserted.connect(self._expand_tree_after_table_insert)
        self._preview_table_model = PreviewTableModel()
        self._stamps = {}
        self._writer_sender = Queue()
        self._writer_receiver = Queue()
        self._writer_process = Process(target=write_task_loop, args=(self._writer_receiver, self._writer_sender))
        self._writer_timer = QTimer(window)
        self._writer_timer.setInterval(100)
        self._writer_timer.timeout.connect(self._communicate)
        self._mapping_tables = {}
        self._ui.preview_tree_view.setModel(self._preview_tree_model)
        self._ui.preview_tree_view.selectionModel().currentChanged.connect(self._change_table)
        self._ui.preview_table_view.setModel(self._preview_table_model)
        self._ui.database_url_combo_box.currentTextChanged.connect(self._reload_preview)
        self._ui.live_preview_check_box.stateChanged.connect(self._ui.frame_preview.setEnabled)
        self._ui.live_preview_check_box.clicked.connect(self._reload_preview)
        self._ui.max_preview_rows_spin_box.valueChanged.connect(self._reload_preview)
        self._ui.max_preview_tables_spin_box.valueChanged.connect(self._reload_preview)
        self._ui.load_url_from_fs_button.clicked.connect(self._load_url_from_filesystem)
        self._ui.mappings_table.selectionModel().currentChanged.connect(self._change_selected_table)
        self._enable_controls()
        if self._url_model.rowCount() > 0:
            self._reload_preview()

    @Slot()
    def _reload_preview(self):
        """Sets the current url and reloads preview."""
        current_url = self._ui.database_url_combo_box.currentText()
        if not current_url:
            return
        self._current_url = current_url
        self._stamps.clear()
        for row in range(self._mappings_table_model.rowCount()):
            index = self._mappings_table_model.index(row, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked:
                self._load_preview_data(index.data())

    @Slot(QModelIndex, QModelIndex)
    def _change_table(self, current, previous):
        """
        Changes preview table data.

        Args:
            current (QModelIndex): index to the currently selected table on the list
            previous (QModelIndex): index to the previously selected table on the list
        """
        current_parent = self._preview_tree_model.parent(current)
        if not current_parent.isValid():
            current_parent = current
        previous_parent = self._preview_tree_model.parent(previous)
        if not previous_parent.isValid():
            previous_parent = previous
        if current_parent != previous_parent:
            name = self._preview_tree_model.data(current_parent)
            index = self._mappings_proxy_model.mapFromSource(self._mappings_table_model.index_of(name))
            self._ui.mappings_table.selectionModel().currentChanged.disconnect(self._change_selected_table)
            self._ui.mappings_table.setCurrentIndex(index)
            self._ui.mappings_table.selectionModel().currentChanged.connect(self._change_selected_table)
        table = current.data(PreviewTreeModel.TABLE_ROLE)
        if table is not None:
            table_name = current.data()
            mapping_name = current.parent().data()
            mapping_colors = self._mapping_editor_table_model.mapping_colors()
            self._preview_table_model.reset(mapping_name, table_name, table, mapping_colors)
        else:
            self._preview_table_model.clear()

    @Slot(QModelIndex, QModelIndex)
    def _change_selected_table(self, current, previous):
        """
        Changes selected table on the tree view after user has selected another mapping.

        Args:
            current (QModelIndex): index to selected mapping on mapping list
            previous (QModelIndex): index to previously selected mapping on mapping list
        """
        if not current.isValid():
            return
        current = self._mappings_proxy_model.mapToSource(current)
        mappings_index = self._mappings_table_model.index(current.row(), 0)
        if mappings_index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        current_preview = self._ui.preview_tree_view.selectionModel().currentIndex()
        if current_preview.parent().row() >= self._preview_tree_model.rowCount():
            # May happen when mappings at the end of the list are removed.
            return
        mapping_name = mappings_index.data()
        if current_preview.parent().data() == mapping_name:
            return
        self._set_current_table(mapping_name)

    def _set_current_table(self, name):
        """Sets the current index of preview tree.

        Args:
            name (str): mapping's name
        """
        for row in range(self._preview_tree_model.rowCount()):
            index = self._preview_tree_model.index(row, 0)
            if index.data() == name:
                if self._preview_tree_model.rowCount(index) > 0:
                    table_index = self._preview_tree_model.index(0, 0, index)
                    self._ui.preview_tree_view.selectionModel().setCurrentIndex(
                        table_index, QItemSelectionModel.ClearAndSelect
                    )
                    return
                self._ui.preview_tree_view.selectionModel().setCurrentIndex(index, QItemSelectionModel.ClearAndSelect)
                return

    @Slot(QModelIndex, QModelIndex, list)
    def _expand_tree_after_table_change(self, top_left, bottom_right, roles):
        """
        Expands preview tree's root items if their tables have received updates.

        Args:
            top_left (QModelIndex): top left corner of changed preview tree data
            bottom_right (QModelIndex): bottom right corner of changed preview tree data
            roles (list of int): roles
        """
        if PreviewTreeModel.TABLE_ROLE not in roles:
            return
        root_index = top_left.parent()
        if not root_index.isValid():
            return
        self._ui.preview_tree_view.setExpanded(root_index, True)

    @Slot(QModelIndex, QModelIndex, list)
    def _expand_tree_after_table_insert(self, parent, first, last):
        """
        Expands preview tree's root items if tables have been added/removed.

        Args:
            parent (QModelIndex): parent index of items that have been added/removed
            first (int): first row index of affected items
            last (int): last row index of affected items
        """
        if not parent.isValid():
            return
        self._ui.preview_tree_view.setExpanded(parent, True)

    @Slot(QModelIndex, QModelIndex, list)
    def _update_table(self, top_left, bottom_right, roles):
        """
        Updates the preview table if its data has changed.

        Args:
            top_left (QModelIndex): top left corner of changed preview tree data
            bottom_right (QModelIndex): bottom right corner of changed preview tree data
            roles (list of int): roles
        """
        if PreviewTreeModel.TABLE_ROLE not in roles:
            return
        current_parent = top_left.parent()
        current_table_name = self._preview_table_model.table_name()
        current_mapping_name = self._preview_table_model.mapping_name()
        for row in range(top_left.row(), bottom_right.row() + 1):
            index = self._preview_tree_model.index(row, 0, current_parent)
            table_name = index.data()
            mapping_name = index.parent().data()
            if current_table_name != table_name or current_mapping_name != mapping_name:
                continue
            table = index.data(PreviewTreeModel.TABLE_ROLE)
            mapping_colors = self._mapping_editor_table_model.mapping_colors()
            self._preview_table_model.reset(mapping_name, table_name, table, mapping_colors)
            break

    def _set_expect_removals_and_inserts(self, except_removals):
        """Connects and disconnects signals with regarding mapping removals.

        Args:
            except_removals (bool): True means that a mapping removal is expected
        """
        if except_removals:
            self._mappings_table_model.rowsAboutToBeRemoved.disconnect(self._remove_mappings)
            self._mappings_table_model.rowsInserted.disconnect(self._add_mappings)
        else:
            self._mappings_table_model.rowsAboutToBeRemoved.connect(self._remove_mappings)
            self._mappings_table_model.rowsInserted.connect(self._add_mappings)

    @Slot(int, int, QModelIndex)
    def _remove_mappings(self, parent, first, last):
        """
        Removes mappings from the preview models.

        Args:
            parent (QModelIndex): parent index, ignored
            first (int): first mapping list row to be removed
            last (int): last mapping list row to be removed
        """
        if self._current_url is None:
            return
        for row in range(last, first - 1, -1):
            index = self._mappings_table_model.index(row, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
                continue
            removed_name = index.data()
            self._preview_tree_model.remove_mapping(removed_name)
            self._stamps.pop((self._current_url, removed_name), None)

    @Slot(QModelIndex, QModelIndex, list)
    def _rename_mapping(self, top_left, bottom_right, roles):
        """
        Renames preview models' mappings.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if Qt.ItemDataRole.DisplayRole not in roles or self._current_url is None:
            return
        make_index = self._mappings_table_model.index
        indexes = [make_index(row, 0) for row in range(self._mappings_table_model.rowCount())]
        names = [
            index.data() for index in indexes if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
        ]
        old_name, new_name = self._preview_tree_model.rename_mappings(names)
        if not old_name:
            return
        self._stamps.pop((self._current_url, old_name), None)
        self._set_current_table(new_name)

    @Slot(QModelIndex, QModelIndex, list)
    def _enable_mapping(self, top_left, bottom_right, roles):
        """
        Removes/adds mappings as they are enabled or disabled.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if Qt.ItemDataRole.CheckStateRole.value not in roles or self._current_url is None:
            return
        first = top_left.row()
        last = bottom_right.row()
        for row in range(first, last + 1):
            index = self._mappings_table_model.index(row, 0)
            enabled = index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Checked
            name = index.data()
            if not enabled and self._preview_tree_model.has_name(name):
                self._preview_tree_model.remove_mapping(name)
                self._stamps.pop((self._current_url, name), None)
            elif not self._preview_tree_model.has_name(name):
                self._load_preview_data(name)

    @Slot(QModelIndex, QModelIndex, list)
    def _update_changed_tables(self, top_left, bottom_right, roles):
        """
        Updates mappings whenever the always write header flag or fixed table name changes.

        Args:
            top_left (QModelIndex): top left corner of modified mappings' in mapping list model
            bottom_right (QModelIndex): bottom right corner of modified mappings' in mapping list model
            roles (list of int): changed data's role
        """
        if not {
            MappingsTableModel.ALWAYS_EXPORT_HEADER_ROLE,
            MappingsTableModel.FIXED_TABLE_NAME_ROLE,
            MappingsTableModel.GROUP_FN_ROLE,
        } & set(roles):
            return
        row = self._mappings_proxy_model.mapToSource(self._ui.mappings_table.currentIndex()).row()
        index = self._mappings_table_model.index(row, 0)
        if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        name = index.data()
        self._load_preview_data(name)

    @Slot()
    def _update_current_mappings_tables(self):
        """Reloads mapping that is currently selected on mapping list."""
        if self._current_url is None:
            return
        current_index = self._ui.mappings_table.currentIndex()
        if not current_index.isValid():
            return
        row = self._mappings_proxy_model.mapToSource(current_index).row()
        index = self._mappings_table_model.index(row, 0)
        if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
            return
        mapping_name = index.data()
        self._load_preview_data(mapping_name)

    @Slot(QModelIndex, int, int)
    def _add_mappings(self, parent, first, last):
        """Adds new mapping for preview.

        Args:
            parent (QModelIndex): parent index - unused
            first (int): first row added to mappings table
            last (int): last row added to mappings table
        """
        if self._current_url is None:
            return
        for _ in range(first, last + 1):
            index = self._mappings_table_model.index(first, 0)
            if index.data(Qt.ItemDataRole.CheckStateRole) == Qt.CheckState.Unchecked.value:
                continue
            mapping_name = index.data()
            self._load_preview_data(mapping_name)

    def _expect_mapping_editor_resets(self, expect_resets):
        """Connects and disconnects signals that handle mapping editor resets.

        Args:
            expect_resets (bool): True to except editor resets
        """
        if expect_resets:
            self._mapping_editor_table_model.modelReset.disconnect(self._update_current_mappings_tables)
        else:
            self._mapping_editor_table_model.modelReset.connect(self._update_current_mappings_tables)

    @Slot(QModelIndex, int, int)
    def _enable_controls_after_url_insertion(self, parent, first, last):
        """Enables controls after new preview database URL has been inserted."""
        self._enable_controls()

    @Slot()
    def _enable_controls(self):
        """Enables and disables widgets as needed."""
        urls_available = self._url_model.rowCount() > 0
        self._ui.max_preview_rows_spin_box.setEnabled(urls_available)
        self._ui.max_preview_tables_spin_box.setEnabled(urls_available)

    def _load_preview_data(self, mapping_name):
        """
        Loads preview data from database into the preview tables if the url is set and live previews are enabled.

        Args:
            mapping_name (str): mapping's name
        """
        if self._current_url is None or not self._ui.live_preview_check_box.isChecked() or not mapping_name:
            return
        mapping_spec = self._mappings_table_model.mapping_specification(mapping_name)
        max_tables = self._ui.max_preview_tables_spin_box.value()
        max_rows = self._ui.max_preview_rows_spin_box.value()
        id_ = (self._current_url, mapping_name)
        stamp = monotonic()
        self._stamps[id_] = stamp
        if not self._writer_process.is_alive():
            self._writer_process.start()
        self._writer_sender.put(
            WriteTableTask(
                self._current_url,
                mapping_name,
                stamp,
                deepcopy(mapping_spec.root),
                mapping_spec.always_export_header,
                max_tables,
                max_rows,
                mapping_spec.group_fn,
            )
        )
        if not self._writer_timer.isActive():
            self._writer_timer.start()
        else:
            # This ensures we don't end up in a deadlock where we keep sending
            # new tasks while the pipe is stuck with finished ones.
            self._communicate()

    @Slot()
    def _communicate(self):
        """Periodically communicates with the writer process."""
        self._writer_timer.stop()
        if self._writer_process is None:
            return
        try:
            message = self._writer_receiver.get_nowait()
        except queue.Empty:
            pass
        else:
            self._add_or_update_data(*message)
        if self._stamps:
            self._writer_timer.setInterval(50)
            self._writer_timer.start()
        else:
            self._writer_timer.setInterval(100)

    @Slot(tuple, str, object, float)
    def _add_or_update_data(self, worker_id, mapping_name, data, stamp):
        """
        Sets preview data.

        Args:
            worker_id (tuple): a worker identifier
            mapping_name (str): mapping's name
            data (dict): mapping from table name to table
            stamp (float): task's time stamp
        """
        current_stamp = self._stamps.pop(worker_id, None)
        if stamp != current_stamp:
            if current_stamp is not None:
                self._stamps[worker_id] = current_stamp
            return
        self._preview_tree_model.add_or_update_tables(mapping_name, data)
        mapping_index = self._ui.mappings_table.selectionModel().currentIndex()
        self._change_selected_table(mapping_index, mapping_index)

    @Slot(bool)
    def _load_url_from_filesystem(self, _):
        """Shows a file dialog and opens selected database in the preview widget."""
        path = self._browse_database()
        if not path:
            return
        url = "sqlite:///" + path
        self._url_model.append(url)
        self._ui.database_url_combo_box.setCurrentText(url)

    def _browse_database(self):
        """
        Queries a database file from the user.

        Returns:
            str: path to database file
        """
        return QFileDialog.getOpenFileName(self._window, "Select database", self._project_dir, "sqlite (*.sqlite)")[0]

    @Slot()
    def _forget_url_model(self):
        """Replaces current URL model with a mock one."""
        self._url_model = FullUrlListModel(self._ui.database_url_combo_box)
        self._url_model.rowsInserted.connect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.connect(self._enable_controls)
        self._url_model.destroyed.connect(self._forget_url_model)
        self._ui.database_url_combo_box.setModel(self._url_model)
        self._reload_preview()

    def tear_down(self):
        """Stops the writer process and cleans up."""
        self._writer_sender.put("quit")
        self._writer_timer.stop()
        self._stamps.clear()
        self._url_model.rowsInserted.disconnect(self._enable_controls_after_url_insertion)
        self._url_model.modelReset.disconnect(self._enable_controls)
        self._url_model.destroyed.disconnect(self._forget_url_model)
        if self._writer_process.is_alive():
            finished = False
            while not finished:
                # Drain the pipe, otherwise writer process's send() may block and the join() below hangs.
                finished = self._writer_receiver.get() == "finished"
            self._writer_process.join()
        self._writer_process = None


@dataclass(frozen=True)
class WriteTableTask:
    url: str
    mapping_name: str
    stamp: float
    mapping: ExportMapping
    always_export_header: bool
    max_tables: int
    max_rows: int
    group_fn: GroupFunction


def write_task_loop(sender, receiver):
    """A task loop that processes table writing tasks.

    Args:
        sender (Queue): Sending queue to communicate with the loop
        receiver (Queue): Receiving queue endpoint to communicate with the loop
    """
    db_map = None
    tasks = []
    try:
        while True:
            if not receiver.empty() or not tasks:
                while True:
                    task = receiver.get()
                    if task == "quit":
                        return
                    tasks.append(task)
                    if receiver.empty():
                        break
            next_task = tasks.pop(0)
            try:
                if db_map is None or next_task.url != db_map.db_url:
                    db_map = DatabaseMapping(next_task.url)
                tables = _write_tables(db_map, next_task)
            except SpineDBVersionError:
                tables = {"error": [["unsupported database version"]]}
            except SpineDBAPIError as error:
                tables = {"error": [[str(error)]]}
            sender.put(((next_task.url, next_task.mapping_name), next_task.mapping_name, tables, next_task.stamp))
    finally:
        sender.put_nowait("finished")


def _write_tables(db_map, write_task):
    """Creates preview tables with given parameters.

    Args:
        db_map (DatabaseMapping): database mapping
        write_task (WriteTableTask): task parameters

    Returns:
        tuple: mappings from table names to tables (list of rows)
    """
    writer = TableWriter()
    write(
        db_map,
        writer,
        write_task.mapping,
        empty_data_header=write_task.always_export_header,
        max_tables=write_task.max_tables,
        max_rows=write_task.max_rows,
        group_fns=write_task.group_fn,
    )
    return writer.tables



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""ImportMappings widget."""
import pickle
from PySide6.QtCore import QItemSelection, QItemSelectionModel, QModelIndex, QPoint, Slot
from PySide6.QtWidgets import QApplication, QHeaderView, QStyledItemDelegate
from spinetoolbox.widgets.custom_delegates import ComboBoxDelegate
from spinetoolbox.widgets.parameter_value_editor import ParameterValueEditor
from ...widgets import combo_box_width
from ..commands import CreateMapping, DeleteMapping, DuplicateMapping, PasteMappings
from ..mvcmodels.mappings_model import FlattenedColumn
from .custom_menus import MappingListMenu
from .filter_edit_delegate import FilterEditDelegate
from .mime_types import MAPPING_LIST_MIME_TYPE

SOURCE_TYPES = ("Constant", "Column", "Row", "Column Header", "Headers", "Table Name", "Mapping Name", "Fixed", "None")


class ImportMappings:
    """
    Controls the 'Mappings' and 'Mapping specifications' part of the window.
    """

    def __init__(self, mappings_model, ui, undo_stack, editor_window):
        """
        Args:
            mappings_model (MappingsModel): mappings model
            ui (Any): import editor window's UI
            undo_stack (QUndoStack): undo stack
            editor_window (ImportEditorWindow): import editor window
        """
        self._editor_window = editor_window
        self._ui = ui
        self._mappings_model = mappings_model
        self._undo_stack = undo_stack
        # initialize interface
        # NOTE: We make the delegate an attribute so it's never accidentally gc'ed
        self._src_type_delegate = ComboBoxDelegate(SOURCE_TYPES)
        self._parameter_constant_value_delegate = ParameterConstantValueDelegate(self._editor_window)
        self._filter_edit_delegate = FilterEditDelegate(self._editor_window)
        self._ui.mapping_spec_table.setItemDelegateForColumn(FlattenedColumn.POSITION_TYPE, self._src_type_delegate)
        self._ui.mapping_spec_table.setItemDelegateForColumn(
            FlattenedColumn.POSITION, self._parameter_constant_value_delegate
        )
        self._ui.mapping_spec_table.setItemDelegateForColumn(FlattenedColumn.REGEXP, self._filter_edit_delegate)
        self._ui.mapping_spec_table.horizontalHeader().sectionCountChanged.connect(self._resize_spec_table_columns)
        # connect signals
        self._mappings_model.modelReset.connect(self._dummify_root_indexes)
        self._mappings_model.dataChanged.connect(self._select_changed_mapping)
        self._mappings_model.rowsInserted.connect(self._update_current_after_mapping_insertion)
        self._mappings_model.rowsInserted.connect(self._set_mapping_list_buttons_enabled)
        self._mappings_model.rowsRemoved.connect(self._show_list_after_mapping_removal)
        self._mappings_model.rowsRemoved.connect(self._set_mapping_list_buttons_enabled)
        self._ui.source_list.selectionModel().currentChanged.connect(self._change_list)
        self._ui.mapping_list.selectionModel().currentChanged.connect(self._change_flattened_mappings)
        self._ui.mapping_list.selectionModel().selectionChanged.connect(self._update_buttons_enabled)
        self._ui.new_button.clicked.connect(self._new_mapping)
        self._ui.remove_button.clicked.connect(self._delete_selected_mapping)
        self._ui.duplicate_button.clicked.connect(self._duplicate_selected_mapping)
        self._ui.mapping_list.customContextMenuRequested.connect(self._show_mapping_list_context_menu)

    def specification_saved(self):
        """Notifies mappings model items that they have been stored into the specification."""
        self._mappings_model.set_source_table_items_into_specification()

    @Slot(QModelIndex, QModelIndex)
    def _change_list(self, current, previous):
        """Loads current source table's mapping list.

        Args:
            current (QModelIndex): currently selected source table index
            previous (QModelIndex): previously selected source table index
        """
        if not current.isValid():
            self._ui.mapping_list.setRootIndex(self._mappings_model.dummy_parent())
            self._ui.mapping_list.selectionModel().setCurrentIndex(QModelIndex(), QItemSelectionModel.ClearAndSelect)
            return
        self._ui.mapping_list.setRootIndex(current)
        if self._mappings_model.rowCount(current) > 0:
            self._ui.mapping_list.selectionModel().setCurrentIndex(
                self._mappings_model.index(0, 0, current), QItemSelectionModel.ClearAndSelect
            )
        else:
            self._ui.mapping_list.selectionModel().setCurrentIndex(QModelIndex(), QItemSelectionModel.ClearAndSelect)
        self._set_mapping_list_buttons_enabled(current)

    @Slot(QModelIndex, int, int)
    def _update_current_after_mapping_insertion(self, table_index, first, last):
        """Selects newly inserted rows in mapping list view.

        Args:
            table_index (QModelIndex): table index
            first (int): first inserted row index
            last (int): last inserted row index
        """
        if not self._mappings_model.is_table_index(table_index):
            return
        if table_index != self._ui.mapping_list.rootIndex():
            self._ui.source_list.selectionModel().setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)
        top_left = self._mappings_model.index(first, 0, table_index)
        bottom_right = self._mappings_model.index(last, 0, table_index)
        selection_model = self._ui.mapping_list.selectionModel()
        selection_model.select(QItemSelection(top_left, bottom_right), QItemSelectionModel.ClearAndSelect)
        selection_model.setCurrentIndex(self._mappings_model.index(last, 0, table_index), QItemSelectionModel.Select)

    @Slot(QModelIndex, int, int)
    def _show_list_after_mapping_removal(self, table_index, first, last):
        """Selects correct mapping list after new mappings have been inserted.

        Args:
            table_index (QModelIndex): table index
            first (int): first inserted row index
            last (int): last inserted row index
        """
        if not self._mappings_model.is_table_index(table_index):
            return
        if table_index != self._ui.mapping_list.rootIndex():
            self._ui.source_list.selectionModel().setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)

    @Slot(QModelIndex)
    def _set_mapping_list_buttons_enabled(self, table_index):
        """Sets New, Duplicate and Remove button enabled state when selecting a source table.

        Args:
            table_index (QModelIndex): table index
        """
        if not table_index.isValid() or self._mappings_model.is_mapping_list_index(table_index):
            return
        number_of_sources = len(self._ui.mapping_list.model())
        number_of_tables = len(self._mappings_model.real_table_names())
        selected_source = table_index.row()
        if table_index.row() == 1 and number_of_tables == 1:
            is_first_or_last = False
        elif number_of_sources - number_of_tables == 1 and selected_source == number_of_sources - 1:
            is_first_or_last = False
        else:
            is_first_or_last = selected_source in [0, number_of_sources - 1]
        self._ui.new_button.setEnabled(not is_first_or_last)

        has_entries = self._mappings_model.rowCount(table_index) > 0
        self._ui.remove_button.setEnabled(has_entries)
        self._ui.duplicate_button.setEnabled(has_entries)

    @Slot(QItemSelection, QItemSelection)
    def _update_buttons_enabled(self, _selected, _deselected):
        enabled = self._ui.mapping_list.selectionModel().hasSelection()
        self._ui.remove_button.setEnabled(enabled)
        self._ui.duplicate_button.setEnabled(enabled)

    @Slot(QModelIndex, QModelIndex)
    def _change_flattened_mappings(self, current, previous):
        """Loads current mapping to component editor.

        Args:
            current (QModelIndex): currently selected mapping list index
            previous (QModelIndex): previously selected mapping list index
        """
        self._ui.mapping_list.selectionModel().select(current, QItemSelectionModel.Select)
        if not current.isValid():
            self._ui.mapping_spec_table.setRootIndex(self._mappings_model.dummy_parent())
            return
        self._ui.mapping_spec_table.setRootIndex(current)

    @Slot()
    def _dummify_root_indexes(self):
        """Makes sure we don't show source table list in other widgets."""
        self._ui.mapping_list.setRootIndex(self._mappings_model.dummy_parent())
        self._ui.mapping_spec_table.setRootIndex(self._mappings_model.dummy_parent())

    @Slot(QModelIndex, QModelIndex, list)
    def _select_changed_mapping(self, top_left, bottom_right, roles):
        """Sets current indexes such that the previously changed mapping is shown on the window.

        Args:
            top_left (QModelIndex): top left index of changes
            bottom_right (QModelIndex): bottom right index of changes
            roles (list of int): Qt's data roles
        """
        if self._mappings_model.is_table_index(top_left):
            return
        if self._mappings_model.is_mapping_list_index(top_left):
            list_selection_model = self._ui.mapping_list.selectionModel()
            list_index = list_selection_model.currentIndex()
            list_selection = QItemSelection(top_left, bottom_right)
            if list_selection.contains(list_index):
                return
            table_index = top_left.parent()
            table_selection_model = self._ui.source_list.selectionModel()
            if table_index != table_selection_model.currentIndex():
                table_selection_model.setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)
            list_selection_model.select(list_selection, QItemSelectionModel.ClearAndSelect)
            list_selection_model.setCurrentIndex(bottom_right, QItemSelectionModel.Select)
            return
        list_index = top_left.parent()
        table_index = list_index.parent()
        table_selection_model = self._ui.source_list.selectionModel()
        if table_index != table_selection_model.currentIndex():
            table_selection_model.setCurrentIndex(table_index, QItemSelectionModel.ClearAndSelect)
        list_selection_model = self._ui.mapping_list.selectionModel()
        if list_index != list_selection_model.currentIndex():
            list_selection_model.setCurrentIndex(list_index, QItemSelectionModel.ClearAndSelect)

    @Slot()
    def _new_mapping(self):
        """
        Pushes a CreateMapping command to the undo stack
        """
        table_index = self._ui.mapping_list.rootIndex()
        list_row = self._mappings_model.rowCount(table_index)
        command = CreateMapping(table_index.row(), self._mappings_model, list_row)
        self._undo_stack.push(command)

    @Slot()
    def _duplicate_selected_mapping(self):
        """
        Pushes a DuplicateMapping command to the undo stack.
        """
        selection_model = self._ui.mapping_list.selectionModel()
        if not selection_model.hasSelection():
            return
        indexes = selection_model.selectedIndexes()
        table_row = indexes[0].parent().row()
        list_rows = sorted(i.row() for i in indexes)
        self._undo_stack.beginMacro("duplicate mapping(s)")
        for list_row in reversed(list_rows):
            self._undo_stack.push(DuplicateMapping(table_row, self._mappings_model, list_row))
        self._undo_stack.endMacro()

    @Slot()
    def _delete_selected_mapping(self):
        """
        Pushes a DeleteMapping command to the undo stack.
        """
        selection_model = self._ui.mapping_list.selectionModel()
        if not selection_model.hasSelection():
            return
        indexes = selection_model.selectedIndexes()
        if len(indexes) == 1:
            index = indexes[0]
            self._undo_stack.push(DeleteMapping(index.parent().row(), self._mappings_model, index.row()))
        else:
            table_row = indexes[0].parent().row()
            rows = sorted(i.row() for i in indexes)
            self._undo_stack.beginMacro("remove mappings")
            for row in reversed(rows):
                self._undo_stack.push(DeleteMapping(table_row, self._mappings_model, row))
            self._undo_stack.endMacro()

    @Slot(QPoint)
    def _show_mapping_list_context_menu(self, pos):
        global_pos = self._ui.mapping_list.mapToGlobal(pos)
        indexes = self._ui.mapping_list.selectionModel().selectedRows()
        source_list_menu = MappingListMenu(self._ui.source_list, global_pos, bool(indexes), self._can_paste_mappings())
        option = source_list_menu.get_action()
        source_list_menu.deleteLater()
        if option == "Copy mapping(s)":
            self._copy_selected_mappings_to_clipboard()
            return
        if option == "Paste mapping(s)":
            table_index = self._ui.source_list.selectionModel().currentIndex()
            index = self._ui.mapping_list.selectionModel().currentIndex()
            target_row = index.row() + 1 if self._mappings_model.rowCount(table_index) > 0 else 0
            clipboard = QApplication.clipboard()
            mime_data = clipboard.mimeData()
            mapping_list_data = pickle.loads(mime_data.data(MAPPING_LIST_MIME_TYPE))
            self._undo_stack.push(PasteMappings(table_index.row(), target_row, self._mappings_model, mapping_list_data))

    def _copy_selected_mappings_to_clipboard(self):
        """Copies selected mappings to system clipboard."""
        selection_model = self._ui.mapping_list.selectionModel()
        if not selection_model.hasSelection():
            return
        indexes = selection_model.selectedIndexes()
        mime_data = self._mappings_model.mimeData(indexes)
        clipboard = QApplication.clipboard()
        clipboard.setMimeData(mime_data)

    def _can_paste_mappings(self):
        """Checks if it is possible to paste mappings.

        Returns:
            bool: True if mappings can be pasted, False otherwise
        """
        has_table = self._ui.source_list.selectionModel().currentIndex().isValid()
        has_target = self._ui.mapping_list.selectionModel().currentIndex().isValid()
        clipboard = QApplication.clipboard()
        mime_data = clipboard.mimeData()
        return mime_data.hasFormat(MAPPING_LIST_MIME_TYPE) and has_table and has_target

    @Slot(int, int)
    def _resize_spec_table_columns(self, old_column_count, new_column_count):
        """Configures mapping specification table's column widths.

        Args:
            old_column_count (int): previous column count
            new_column_count (int): current column count
        """
        if new_column_count != len(FlattenedColumn):
            return
        spec_table_header = self._ui.mapping_spec_table.horizontalHeader()
        spec_table_header.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)
        spec_table_header.setSectionResizeMode(FlattenedColumn.POSITION_TYPE, QHeaderView.ResizeMode.Fixed)
        spec_table_header.resizeSection(
            FlattenedColumn.POSITION_TYPE, combo_box_width(self._editor_window, SOURCE_TYPES)
        )


class ParameterConstantValueDelegate(QStyledItemDelegate):
    """A delegate that shows a ParameterValueEditor for constant value mappings."""

    def __init__(self, parent):
        super().__init__(parent)

    def createEditor(self, parent, option, index):
        if index.column() != 2:
            return super().createEditor(parent, option, index)
        target = index.siblingAtColumn(0).data()
        ref_type = index.siblingAtColumn(1).data()
        if target.endswith("values") and ref_type == "Constant":
            editor = ParameterValueEditor(index, parent)  # TODO: plain=True for parameter value lists
            editor.show()
            return None
        return super().createEditor(parent, option, index)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains Data Connection's executable item as well as support utilities."""
import os
from spine_engine.project_item.executable_item_base import ExecutableItemBase
from spine_engine.utils.serialization import deserialize_path
from .item_info import ItemInfo
from .output_resources import scan_for_resources
from .utils import restore_database_references


class ExecutableItem(ExecutableItemBase):
    """The executable parts of Data Connection."""

    def __init__(self, name, file_references, db_references, project_dir, logger):
        """
        Args:
            name (str): item's name
            file_references (list): a list of absolute paths to connected files
            db_references (list): a list of url dicts to connected dbs
            project_dir (str): absolute path to project directory
            logger (LoggerInterface): a logger
        """
        super().__init__(name, project_dir, logger)
        data_files = []
        with os.scandir(self._data_dir) as scan_iterator:
            for entry in scan_iterator:
                if entry.is_file():
                    data_files.append(entry.path)
        self._file_paths = file_references + data_files
        self._urls = db_references

    @staticmethod
    def item_type():
        """Returns DataConnectionExecutable's type identifier string."""
        return ItemInfo.item_type()

    def _output_resources_forward(self):
        """See base class."""
        return scan_for_resources(self, self._file_paths, self._urls, self._project_dir)

    @classmethod
    def from_dict(cls, item_dict, name, project_dir, app_settings, specifications, logger):
        """See base class."""
        file_references = item_dict["file_references"]
        file_references = [deserialize_path(r, project_dir) for r in file_references]
        db_references = restore_database_references(
            item_dict.get("db_references", []), item_dict.get("db_credentials", {}), project_dir
        )
        return cls(name, file_references, db_references, project_dir, logger)



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains shared functions that provide copy-paste functionality."""
import pickle
from PySide6.QtCore import QMimeData
from PySide6.QtWidgets import QApplication
from ..commands import InsertRow, SetData


def copy_table_data(table_view, mime_type):
    """Copies data to system clipboard.

    Args:
        table_view (QTableView): target table view
        mime_type (str): mime type for table's native data
    """
    indexes = table_view.selectionModel().selectedIndexes()
    if not indexes:
        return
    rows = [i.row() for i in indexes]
    min_row = min(rows)
    max_row = max(rows)
    columns = [i.column() for i in indexes]
    min_column = min(columns)
    max_column = max(columns)
    index_lookup = {(i.row(), i.column()): i for i in indexes}
    model = table_view.model().sourceModel()
    data = []
    for row in range(min_row, max_row + 1):
        row_data = []
        for column in range(min_column, max_column + 1):
            index = index_lookup.get((row, column))
            if index is None:
                row_data.append(None)
            else:
                row_data.append(index.data(model.GET_DATA_ROLES[column]))
        data.append(row_data)
    mime_data = QMimeData()
    mime_data.setData(mime_type, pickle.dumps(data))
    clipboard = QApplication.clipboard()
    clipboard.setMimeData(mime_data)


def paste_table_data(table_view, mime_type, undo_stack):
    """Pushes commands that Paste data from system clipboard to undo stack.

    Args:
        table_view (QTableView): target table view
        mime_type (str): mime type for table's native data
        undo_stack (QUndoStack): undo stack
    """
    if not table_view.selectionModel().hasSelection():
        return
    clipboard = QApplication.clipboard()
    mime_data = clipboard.mimeData()
    if mime_data is None:
        return
    if not mime_data.hasFormat(mime_type):
        return
    data = pickle.loads(mime_data.data(mime_type))
    indexes = table_view.selectionModel().selectedIndexes()
    model = table_view.model()
    source_model = model.sourceModel()
    model.setDynamicSortFilter(False)
    undo_stack.beginMacro("paste data")
    if len(indexes) == 1:
        _paste_all(indexes[0], data, model, source_model, undo_stack)
    else:
        _paste_fit_to_selection(indexes, data, model, source_model, undo_stack)
    model.setDynamicSortFilter(True)
    undo_stack.endMacro()


def _paste_all(top_left, data, model, source_model, undo_stack):
    """Pushes commands that paste all data on clipboard extending target model if needed to undo stack.

    Args:
        top_left (QModelIndex): top left corner of target area
        data (list of list): rows to paste
        model (QSortFilterModel): sorted model
        source_model (QAbstractTableModel): actual model
        undo_stack (QUndoStack): undo stack
    """
    for row, row_data in enumerate(data):
        table_row = row + top_left.row()
        if table_row >= model.rowCount():
            undo_stack.push(InsertRow("", source_model, table_row, None))
        for column, item in enumerate(row_data):
            table_column = column + top_left.column()
            if item is None or table_column >= model.columnCount():
                continue
            index = model.index(table_row, table_column)
            undo_stack.push(
                SetData(
                    "",
                    index,
                    item,
                    index.data(source_model.GET_DATA_ROLES[table_column]),
                    source_model.SET_DATA_ROLES[table_column],
                )
            )


def _paste_fit_to_selection(indexes, data, model, source_model, undo_stack):
    """Pushes commands that paste data on clipboard to undo stack.

    Args:
        indexes (QModelIndex): selected indexes
        data (list of list): rows to paste
        model (QSortFilterModel): sorted model
        source_model (QAbstractTableModel): actual model
        undo_stack (QUndoStack): undo stack
    """
    rows = [i.row() for i in indexes]
    min_row = min(rows)
    max_row = min(max(rows), min_row + len(data) - 1)
    columns = [i.column() for i in indexes]
    min_column = min(columns)
    max_column = min(max(columns), min_column + len(data[0]) - 1)
    for row in range(min_row, max_row + 1):
        row_data = data[row - min_row]
        for column in range(min_column, max_column + 1):
            if column >= model.columnCount():
                continue
            item = row_data[column - min_column]
            if item is None:
                continue
            index = model.index(row, column)
            undo_stack.push(
                SetData(
                    "",
                    index,
                    item,
                    index.data(source_model.GET_DATA_ROLES[column]),
                    source_model.SET_DATA_ROLES[column],
                )
            )



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Toolbox.
# Spine Toolbox is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains unit tests for the tool_instance module."""
from pathlib import Path
import sys
from tempfile import TemporaryDirectory
import unittest
from unittest import mock
from spine_items.tool.tool_specifications import ExecutableTool, GAMSTool, JuliaTool, PythonTool
from tests.mock_helpers import MockQSettings


class TestToolInstance(unittest.TestCase):
    def test_python_prepare_with_jupyter_console(self):
        # No cmd line args
        instance = self._make_python_tool_instance(True)
        with mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_manager:
            instance.prepare([])
            mock_manager.assert_called_once()
            self.assertEqual(mock_manager.call_args[0][1], "some_kernel")  # kernel_name
            self.assertEqual(mock_manager.call_args[0][2], ["%cd -q path/", '%run "main.py"'])  # commands
        # With tool cmd line args
        instance = self._make_python_tool_instance(True)
        with mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_manager:
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called_once()
            self.assertEqual(mock_manager.call_args[0][1], "some_kernel")
            self.assertEqual(mock_manager.call_args[0][2], ["%cd -q path/", '%run "main.py" "arg1" "arg2"'])
        # With tool and tool spec cmd line args
        instance = self._make_python_tool_instance(True, ["arg3"])
        with mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_manager:
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called_once()
            self.assertEqual(mock_manager.call_args[0][1], "some_kernel")
            self.assertEqual(mock_manager.call_args[0][2], ["%cd -q path/", '%run "main.py" "arg3" "arg1" "arg2"'])

    def test_python_prepare_with_basic_console(self):
        # No cmd line args
        instance = self._make_python_tool_instance(False)
        with mock.patch("spine_items.tool.tool_instance.PythonPersistentExecutionManager") as mock_manager:
            mock_manager.return_value = True
            instance.prepare([])
            mock_manager.assert_called_once()
            self.assertEqual([sys.executable], mock_manager.call_args[0][1])  # args
            self.assertEqual(5, len(mock_manager.call_args[0][2]))  # commands
            self.assertEqual("python main.py", mock_manager.call_args[0][3])  # alias
        # With tool cmd line args
        instance = self._make_python_tool_instance(False)
        with mock.patch("spine_items.tool.tool_instance.PythonPersistentExecutionManager") as mock_manager:
            mock_manager.return_value = True
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called_once()
            self.assertEqual([sys.executable], mock_manager.call_args[0][1])  # args
            self.assertEqual(5, len(mock_manager.call_args[0][2]))  # commands
            self.assertEqual("python main.py arg1 arg2", mock_manager.call_args[0][3])  # alias
        # With tool and tool spec cmd line args
        instance = self._make_python_tool_instance(False, ["arg3"])
        with mock.patch("spine_items.tool.tool_instance.PythonPersistentExecutionManager") as mock_manager:
            mock_manager.return_value = True
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called_once()
            self.assertEqual([sys.executable], mock_manager.call_args[0][1])  # args
            self.assertEqual(5, len(mock_manager.call_args[0][2]))  # commands
            self.assertEqual("python main.py arg3 arg1 arg2", mock_manager.call_args[0][3])  # alias

    def test_julia_prepare_with_jupyter_console(self):
        # No cmd line args
        instance = self._make_julia_tool_instance(True)
        with (
            mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_kem,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.custom_find_kernel_specs") as mock_find_kernel_specs,
        ):
            mock_find_kernel_specs.return_value = {"some_julia_kernel": Path(__file__).parent / "dummy_julia_kernel"}
            mock_isfile.return_value = False
            instance.prepare([])
            mock_isfile.assert_called()
            mock_kem.assert_called_once()
            mock_find_kernel_specs.assert_called_once()
            self.assertEqual(mock_kem.call_args[0][1], "some_julia_kernel")  # kernel_name
            self.assertEqual(mock_kem.call_args[0][2], ['cd("path/");', 'include("hello.jl")'])  # commands
        # With tool cmd line args
        instance = self._make_julia_tool_instance(True)
        with (
            mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_kem,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.custom_find_kernel_specs") as mock_find_kernel_specs,
        ):
            mock_find_kernel_specs.return_value = {"some_julia_kernel": Path(__file__).parent / "dummy_julia_kernel"}
            mock_isfile.return_value = False
            instance.prepare(["arg1", "arg2"])
            mock_isfile.assert_called()
            mock_kem.assert_called_once()
            mock_find_kernel_specs.assert_called_once()
            self.assertEqual(mock_kem.call_args[0][1], "some_julia_kernel")
            self.assertEqual(
                mock_kem.call_args[0][2],
                ['cd("path/");', 'empty!(ARGS); append!(ARGS, ["arg1", "arg2"]);', 'include("hello.jl")'],
            )
        # With tool and tool spec cmd line args
        instance = self._make_julia_tool_instance(True, ["arg3"])
        with (
            mock.patch("spine_items.tool.tool_instance.KernelExecutionManager") as mock_kem,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.custom_find_kernel_specs") as mock_find_kernel_specs,
        ):
            mock_find_kernel_specs.return_value = {"some_julia_kernel": Path(__file__).parent / "dummy_julia_kernel"}
            mock_isfile.return_value = False
            instance.prepare(["arg1", "arg2"])
            mock_isfile.assert_called()
            mock_kem.assert_called_once()
            mock_find_kernel_specs.assert_called_once()
            self.assertEqual(mock_kem.call_args[0][1], "some_julia_kernel")
            self.assertEqual(
                mock_kem.call_args[0][2],
                ['cd("path/");', 'empty!(ARGS); append!(ARGS, ["arg3", "arg1", "arg2"]);', 'include("hello.jl")'],
            )

    def test_julia_prepare_with_basic_console(self):
        # No cmd line args
        instance = self._make_julia_tool_instance(False)
        instance._owner.options = {"julia_sysimage": "path/to/sysimage.so"}
        with (
            mock.patch("spine_items.tool.tool_instance.JuliaPersistentExecutionManager") as mock_manager,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.resolve_julia_executable") as mock_resolve_julia,
        ):
            mock_isfile.return_value = True  # Make isfile() accept fake julia_sysimage path
            mock_manager.return_value = True
            mock_resolve_julia.return_value = "path/to/julia"
            instance.prepare([])
            mock_isfile.assert_called()
            mock_manager.assert_called_once()
            mock_resolve_julia.assert_called()
            self.assertEqual(
                ["path/to/julia", "--sysimage=path/to/sysimage.so"], mock_manager.call_args[0][1]
            )  # args attribute for JuliaPersistentExecutionManger
            self.assertEqual(['cd("path/");', 'include("hello.jl")'], mock_manager.call_args[0][2])  # commands
            self.assertEqual("julia hello.jl", mock_manager.call_args[0][3])  # alias
        # With tool cmd line args
        instance = self._make_julia_tool_instance(False)
        with (
            mock.patch("spine_items.tool.tool_instance.JuliaPersistentExecutionManager") as mock_manager,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.resolve_julia_executable") as mock_resolve_julia,
        ):
            mock_isfile.return_value = False
            mock_manager.return_value = True
            mock_resolve_julia.return_value = "path/to/julia"
            instance.prepare(["arg1", "arg2"])
            mock_isfile.assert_called()
            mock_manager.assert_called_once()
            mock_resolve_julia.assert_called()
            self.assertEqual(["path/to/julia"], mock_manager.call_args[0][1])
            self.assertEqual(
                ['cd("path/");', 'empty!(ARGS); append!(ARGS, ["arg1", "arg2"]);', 'include("hello.jl")'],
                mock_manager.call_args[0][2],
            )
            self.assertEqual("julia hello.jl arg1 arg2", mock_manager.call_args[0][3])  # alias
        # With tool and tool spec cmd line args
        instance = self._make_julia_tool_instance(False, ["arg3"])
        with (
            mock.patch("spine_items.tool.tool_instance.JuliaPersistentExecutionManager") as mock_manager,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("spine_items.tool.utils.resolve_julia_executable") as mock_resolve_julia,
        ):
            mock_isfile.return_value = False
            mock_manager.return_value = True
            mock_resolve_julia.return_value = "path/to/julia"
            instance.prepare(["arg1", "arg2"])
            mock_isfile.assert_called()
            mock_manager.assert_called_once()
            mock_resolve_julia.assert_called()
            self.assertEqual(["path/to/julia"], mock_manager.call_args[0][1])
            self.assertEqual(
                ['cd("path/");', 'empty!(ARGS); append!(ARGS, ["arg3", "arg1", "arg2"]);', 'include("hello.jl")'],
                mock_manager.call_args[0][2],
            )
            self.assertEqual("julia hello.jl arg3 arg1 arg2", mock_manager.call_args[0][3])  # alias

    def test_prepare_sysimg_maker(self):
        instance = self._make_julia_tool_instance(False)
        instance._settings = FakeQSettings()
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_pem,
            mock.patch("spine_items.tool.utils.resolve_julia_executable") as mock_resolve_julia,
        ):
            mock_resolve_julia.return_value = "path/to/julia"
            instance.prepare([])
            mock_pem.assert_called_once()
            mock_resolve_julia.assert_called()
            self.assertEqual("path/to/julia", mock_pem.call_args[0][1])  # Julia exe
            self.assertEqual(["hello.jl"], mock_pem.call_args[0][2])  # script
            self.assertEqual("path/", mock_pem.call_args[1]["workdir"])  # workdir
        instance.terminate_instance()  # Increase coverage

    def test_julia_prepare_with_invalid_kernel(self):
        instance = self._make_julia_tool_instance(True)
        with self.assertRaises(RuntimeError):
            instance.prepare([])
        self.assertEqual(None, instance.exec_mngr)
        self.assertEqual(False, instance.is_running())
        instance.terminate_instance()  # Cover terminate_instance()

    def test_gams_prepare_with_cmd_line_arguments(self):
        # No cmd line args
        instance = self._make_gams_tool_instance()
        path_to_gams = "path/to/gams"
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager,
            mock.patch("spine_items.tool.tool_instance.resolve_gams_executable") as mock_gams_exe,
        ):
            mock_manager.return_value = True
            mock_gams_exe.return_value = path_to_gams
            instance.prepare([])
            mock_manager.assert_called()
            mock_gams_exe.assert_called()
            self.assertEqual(path_to_gams, instance.program)
            self.assertEqual(3, len(instance.args))
            self.assertEqual("model.gms", instance.args[0])
            self.assertEqual("curDir=path/", instance.args[1])
            self.assertEqual("logoption=3", instance.args[2])
        # With tool cmd line args
        instance = self._make_gams_tool_instance()
        path_to_gams = "path/to/gams"
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager,
            mock.patch("spine_items.tool.tool_instance.resolve_gams_executable") as mock_gams_exe,
        ):
            mock_manager.return_value = True
            mock_gams_exe.return_value = path_to_gams
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called()
            mock_gams_exe.assert_called()
            self.assertEqual(path_to_gams, instance.program)
            self.assertEqual(5, len(instance.args))
            self.assertEqual("model.gms", instance.args[0])
            self.assertEqual("curDir=path/", instance.args[1])
            self.assertEqual("logoption=3", instance.args[2])
            self.assertEqual("arg1", instance.args[3])
            self.assertEqual("arg2", instance.args[4])
        # With tool and tool spec cmd line args
        instance = self._make_gams_tool_instance(tool_spec_args=["arg3"])
        path_to_gams = "path/to/gams"
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager,
            mock.patch("spine_items.tool.tool_instance.resolve_gams_executable") as mock_gams_exe,
        ):
            mock_manager.return_value = True
            mock_gams_exe.return_value = path_to_gams
            instance.prepare(["arg1", "arg2"])
            mock_manager.assert_called()
            mock_gams_exe.assert_called()
            self.assertEqual(path_to_gams, instance.program)
            self.assertEqual(6, len(instance.args))
            self.assertEqual("model.gms", instance.args[0])
            self.assertEqual("curDir=path/", instance.args[1])
            self.assertEqual("logoption=3", instance.args[2])
            self.assertEqual("arg3", instance.args[3])
            self.assertEqual("arg1", instance.args[4])
            self.assertEqual("arg2", instance.args[5])

    def test_executable_prepare_with_main_program(self):
        instance = self._make_executable_tool_instance(tool_spec_args=["arg3"])
        # when os.path.isfile fails, we throw a RuntimeError
        self.assertRaises(RuntimeError, instance.prepare, ["arg1", "arg2"])
        # Test when sys.platform is win32
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("sys.platform", "win32"),
        ):
            mock_isfile.return_value = True
            mock_manager.return_value = True
            instance.prepare(["arg1", "arg2"])  # With tool cmd line args
            self.assertEqual(1, mock_manager.call_count)
            self.assertEqual(1, mock_isfile.call_count)
            self.assertEqual("path/program.exe", instance.program)
            self.assertEqual(3, len(instance.args))
            self.assertEqual(["arg3", "arg1", "arg2"], instance.args)
            instance = self._make_executable_tool_instance()
            instance.prepare([])  # Without cmd line args
            self.assertEqual(2, mock_manager.call_count)
            self.assertEqual(2, mock_isfile.call_count)
            self.assertEqual("path/program.exe", instance.program)
            self.assertEqual(0, len(instance.args))
        # Test when sys.platform is linux
        with (
            mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager,
            mock.patch("os.path.isfile") as mock_isfile,
            mock.patch("sys.platform", "linux"),
        ):
            instance = self._make_executable_tool_instance(tool_spec_args=["arg3"])
            mock_isfile.return_value = True
            mock_manager.return_value = True
            instance.prepare(["arg1", "arg2"])  # With cmd line args
            self.assertEqual(1, mock_manager.call_count)
            self.assertEqual(1, mock_isfile.call_count)
            self.assertEqual("sh", instance.program)
            self.assertEqual(4, len(instance.args))
            self.assertEqual(["path/program.exe", "arg3", "arg1", "arg2"], instance.args)

    def test_executable_prepare_with_cmd(self):
        instance = self._make_executable_tool_instance(shell="cmd.exe", cmd="dir", tool_spec_args=["arg3"])
        with mock.patch("spine_items.tool.tool_instance.ProcessExecutionManager") as mock_manager:
            # Run command with cmd.exe
            mock_manager.return_value = True
            instance.prepare(["arg1", "arg2"])
            self.assertEqual(1, mock_manager.call_count)
            self.assertEqual("cmd.exe", instance.program)
            self.assertEqual(5, len(instance.args))
            self.assertEqual(["/C", "dir", "arg3", "arg1", "arg2"], instance.args)
            # Run command with bash shell
            instance = self._make_executable_tool_instance(shell="bash", cmd="ls")
            instance.prepare(["-a"])
            self.assertEqual(2, mock_manager.call_count)
            self.assertEqual("sh", instance.program)
            self.assertEqual(2, len(instance.args))
            self.assertEqual(["ls", "-a"], instance.args)
            # Run command without shell
            instance = self._make_executable_tool_instance(cmd="cat")
            instance.prepare(["file.txt"])
            self.assertEqual(3, mock_manager.call_count)
            self.assertEqual("cat", instance.program)
            self.assertEqual(1, len(instance.args))
            self.assertEqual(["file.txt"], instance.args)

    def test_execute_julia_tool_instance(self):
        instance = self._make_julia_tool_instance(False)
        self.execute_fake_python_julia_and_executable_tool_instances(instance)

    def test_execute_python_tool_instance(self):
        instance = self._make_python_tool_instance(False)
        self.execute_fake_python_julia_and_executable_tool_instances(instance)

    def test_execute_executable_tool_instance(self):
        instance = self._make_executable_tool_instance("cmd.exe", cmd="dir")
        self.execute_fake_python_julia_and_executable_tool_instances(instance)

    def execute_fake_python_julia_and_executable_tool_instances(self, instance):
        """Python and Julia Tool Specification return codes are the same."""
        instance.exec_mngr = FakeExecutionManager(0)  # Valid return code
        self.assertEqual(0, instance.execute())
        instance.exec_mngr = FakeExecutionManager(-1)  # Valid return code
        self.assertEqual(-1, instance.execute())
        instance.exec_mngr = FakeExecutionManager(1)  # Invalid return code
        self.assertEqual(1, instance.execute())

    def test_execute_gams_tool_instance(self):
        temp_dir = TemporaryDirectory()
        instance = self._make_gams_tool_instance(temp_dir.name)
        instance.exec_mngr = FakeExecutionManager(0)  # Valid return code
        self.assertEqual(0, instance.execute())
        instance.exec_mngr = FakeExecutionManager(1)  # Valid return code
        self.assertEqual(1, instance.execute())  # This creates a GAMS project file for debugging
        debug_gpr_path = Path(temp_dir.name) / "specification_name_autocreated.gpr"
        self.assertTrue(debug_gpr_path.is_file())
        debug_gpr_path.unlink()  # Remove file (Path.unlink is equivalent to os.remove)
        self.assertFalse(debug_gpr_path.is_file())
        instance.exec_mngr = FakeExecutionManager(-1)  # Invalid return code
        self.assertEqual(-1, instance.execute())  # This creates a GAMS project file for debugging
        self.assertTrue(debug_gpr_path.is_file())
        temp_dir.cleanup()

    @staticmethod
    def _make_python_tool_instance(use_jupyter_console, tool_spec_args=None):
        specification = PythonTool(
            "specification name",
            "python",
            "",
            ["main.py"],
            MockQSettings(),
            mock.MagicMock(),
            cmdline_args=tool_spec_args,
        )
        specification.init_execution_settings()
        if use_jupyter_console:
            specification.execution_settings["use_jupyter_console"] = True
            specification.execution_settings["kernel_spec_name"] = "some_kernel"
        return specification.create_tool_instance("path/", False, logger=mock.MagicMock(), owner=mock.MagicMock())

    @staticmethod
    def _make_julia_tool_instance(use_jupyter_console, tool_spec_args=None):
        specification = JuliaTool(
            "specification name",
            "julia",
            "",
            ["hello.jl"],
            MockQSettings(),
            mock.MagicMock(),
            cmdline_args=tool_spec_args,
        )
        specification.init_execution_settings()
        if use_jupyter_console:
            specification.execution_settings["use_jupyter_console"] = True
            specification.execution_settings["kernel_spec_name"] = "some_julia_kernel"
        return specification.create_tool_instance("path/", False, logger=mock.MagicMock(), owner=mock.MagicMock())

    @staticmethod
    def _make_gams_tool_instance(temp_dir=None, tool_spec_args=None):
        path = temp_dir if temp_dir else ""
        specification = GAMSTool(
            "specification name",
            "gams",
            path,
            ["model.gms"],
            MockQSettings(),
            mock.MagicMock(),
            cmdline_args=tool_spec_args,
        )
        return specification.create_tool_instance("path/", False, logger=mock.MagicMock(), owner=mock.MagicMock())

    @staticmethod
    def _make_executable_tool_instance(shell=None, cmd=None, tool_spec_args=None):
        if cmd:
            specification = ExecutableTool(
                "name", "executable", "", [], MockQSettings(), mock.MagicMock(), cmdline_args=tool_spec_args
            )
        else:
            specification = ExecutableTool(
                "name",
                "executable",
                "",
                ["program.exe"],
                MockQSettings(),
                mock.MagicMock(),
                cmdline_args=tool_spec_args,
            )
        specification.init_execution_settings()
        if shell == "cmd.exe":
            specification.execution_settings["shell"] = "cmd.exe"
        elif shell == "bash":
            specification.execution_settings["shell"] = "bash"
        if cmd:
            specification.execution_settings["cmd"] = cmd
        return specification.create_tool_instance("path/", False, logger=mock.MagicMock(), owner=mock.MagicMock())


class FakeExecutionManager:
    def __init__(self, retval):
        self.retval = retval

    def run_until_complete(self):
        return self.retval


class FakeQSettings:
    def value(self, key, defaultValue=""):
        if key == "appSettings/makeSysImage":
            return "true"


if __name__ == "__main__":
    unittest.main()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Unit tests for :class:`ClassRenamesTableModel`."""
import unittest
from PySide6.QtCore import Qt
from PySide6.QtGui import QUndoStack
from spine_items.data_transformer.mvcmodels.class_renames_table_model import ClassRenamesTableModel


class TestClassRenamesTableModel(unittest.TestCase):
    def setUp(self):
        self._undo_stack = QUndoStack()

    def tearDown(self):
        self._undo_stack.deleteLater()

    def test_columnCount_always_two(self):
        model = ClassRenamesTableModel(self._undo_stack, {})
        self.assertEqual(model.columnCount(), 2)

    def test_rowCount(self):
        model = ClassRenamesTableModel(self._undo_stack, {})
        self.assertEqual(model.rowCount(), 0)
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A", "b": "B"})
        self.assertEqual(model.rowCount(), 2)

    def test_data(self):
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A", "b": "b"})
        self.assertEqual(model.index(0, 0).data(), "a")
        self.assertEqual(model.index(0, 1).data(), "A")
        self.assertEqual(model.index(1, 0).data(), "b")
        self.assertEqual(model.index(1, 1).data(), "b")

    def test_both_columns_are_editable(self):
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A"})
        self.assertTrue(model.index(0, 0).flags() & Qt.ItemIsEditable)
        self.assertTrue(model.index(0, 1).flags() & Qt.ItemIsEditable)

    def test_headers(self):
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A"})
        self.assertIsNone(model.headerData(0, Qt.Orientation.Vertical))
        self.assertEqual(model.headerData(0, Qt.Orientation.Horizontal), "Original")
        self.assertEqual(model.headerData(1, Qt.Orientation.Horizontal), "Renamed")

    def test_renaming_settings(self):
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A", "b": "b", "c": ""})
        self.assertEqual(model.renaming_settings(), {"a": "A", "b": "b", "c": ""})

    def test_setData(self):
        model = ClassRenamesTableModel(self._undo_stack, {"a": "A"})
        self.assertTrue(model.setData(model.index(0, 1), "B"))
        self.assertEqual(model.renaming_settings(), {"a": "B"})


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

################################################################################
## Form generated from reading UI file 'tool_specification_form.ui'
##
## Created by: Qt User Interface Compiler version 6.7.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide6.QtCore import (QCoreApplication, QDate, QDateTime, QLocale,
    QMetaObject, QObject, QPoint, QRect,
    QSize, QTime, QUrl, Qt)
from PySide6.QtGui import (QAction, QBrush, QColor, QConicalGradient,
    QCursor, QFont, QFontDatabase, QGradient,
    QIcon, QImage, QKeySequence, QLinearGradient,
    QPainter, QPalette, QPixmap, QRadialGradient,
    QTransform)
from PySide6.QtWidgets import (QAbstractItemView, QApplication, QComboBox, QDockWidget,
    QFrame, QHBoxLayout, QHeaderView, QLabel,
    QLineEdit, QMainWindow, QSizePolicy, QStatusBar,
    QVBoxLayout, QWidget)

from spinetoolbox.widgets.code_text_edit import CodeTextEdit
from spinetoolbox.widgets.custom_qtreeview import (CustomTreeView, SourcesTreeView)
from spine_items import resources_icons_rc

class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        if not MainWindow.objectName():
            MainWindow.setObjectName(u"MainWindow")
        MainWindow.resize(991, 643)
        MainWindow.setDockNestingEnabled(True)
        self.actionNew_main_program_file = QAction(MainWindow)
        self.actionNew_main_program_file.setObjectName(u"actionNew_main_program_file")
        icon = QIcon()
        icon.addFile(u":/icons/file-regular.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionNew_main_program_file.setIcon(icon)
        self.actionSelect_main_program_file = QAction(MainWindow)
        self.actionSelect_main_program_file.setObjectName(u"actionSelect_main_program_file")
        icon1 = QIcon()
        icon1.addFile(u":/icons/folder-open-regular.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionSelect_main_program_file.setIcon(icon1)
        self.actionAdd_program_file = QAction(MainWindow)
        self.actionAdd_program_file.setObjectName(u"actionAdd_program_file")
        self.actionAdd_program_file.setEnabled(False)
        self.actionAdd_program_file.setIcon(icon1)
        self.actionAdd_program_directory = QAction(MainWindow)
        self.actionAdd_program_directory.setObjectName(u"actionAdd_program_directory")
        self.actionAdd_program_directory.setEnabled(False)
        icon2 = QIcon()
        icon2.addFile(u":/icons/folder-open-solid.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionAdd_program_directory.setIcon(icon2)
        self.actionRemove_selected_program_files = QAction(MainWindow)
        self.actionRemove_selected_program_files.setObjectName(u"actionRemove_selected_program_files")
        self.actionRemove_selected_program_files.setEnabled(False)
        icon3 = QIcon()
        icon3.addFile(u":/icons/trash-alt.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionRemove_selected_program_files.setIcon(icon3)
        self.actionAdd_input_files = QAction(MainWindow)
        self.actionAdd_input_files.setObjectName(u"actionAdd_input_files")
        icon4 = QIcon()
        icon4.addFile(u":/icons/plus.svg", QSize(), QIcon.Mode.Normal, QIcon.State.Off)
        self.actionAdd_input_files.setIcon(icon4)
        self.actionRemove_selected_input_files = QAction(MainWindow)
        self.actionRemove_selected_input_files.setObjectName(u"actionRemove_selected_input_files")
        self.actionRemove_selected_input_files.setEnabled(False)
        self.actionRemove_selected_input_files.setIcon(icon3)
        self.actionAdd_opt_input_files = QAction(MainWindow)
        self.actionAdd_opt_input_files.setObjectName(u"actionAdd_opt_input_files")
        self.actionAdd_opt_input_files.setIcon(icon4)
        self.actionRemove_selected_opt_input_files = QAction(MainWindow)
        self.actionRemove_selected_opt_input_files.setObjectName(u"actionRemove_selected_opt_input_files")
        self.actionRemove_selected_opt_input_files.setEnabled(False)
        self.actionRemove_selected_opt_input_files.setIcon(icon3)
        self.actionAdd_output_files = QAction(MainWindow)
        self.actionAdd_output_files.setObjectName(u"actionAdd_output_files")
        self.actionAdd_output_files.setIcon(icon4)
        self.actionRemove_selected_output_files = QAction(MainWindow)
        self.actionRemove_selected_output_files.setObjectName(u"actionRemove_selected_output_files")
        self.actionRemove_selected_output_files.setEnabled(False)
        self.actionRemove_selected_output_files.setIcon(icon3)
        self.actionNew_program_file = QAction(MainWindow)
        self.actionNew_program_file.setObjectName(u"actionNew_program_file")
        self.actionNew_program_file.setEnabled(False)
        self.actionNew_program_file.setIcon(icon)
        self.actionSave_program_file = QAction(MainWindow)
        self.actionSave_program_file.setObjectName(u"actionSave_program_file")
        self.actionSave_program_file.setEnabled(False)
        self.actionRemove_all_program_files = QAction(MainWindow)
        self.actionRemove_all_program_files.setObjectName(u"actionRemove_all_program_files")
        self.actionRemove_all_program_files.setIcon(icon3)
        self.centralwidget = QWidget(MainWindow)
        self.centralwidget.setObjectName(u"centralwidget")
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Preferred, QSizePolicy.Policy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.centralwidget.sizePolicy().hasHeightForWidth())
        self.centralwidget.setSizePolicy(sizePolicy)
        self.verticalLayout_3 = QVBoxLayout(self.centralwidget)
        self.verticalLayout_3.setSpacing(6)
        self.verticalLayout_3.setObjectName(u"verticalLayout_3")
        self.verticalLayout_3.setContentsMargins(0, 6, 0, 0)
        self.horizontalLayout_2 = QHBoxLayout()
        self.horizontalLayout_2.setSpacing(9)
        self.horizontalLayout_2.setObjectName(u"horizontalLayout_2")
        self.horizontalLayout_2.setContentsMargins(6, -1, 6, -1)
        self.label_tooltype = QLabel(self.centralwidget)
        self.label_tooltype.setObjectName(u"label_tooltype")

        self.horizontalLayout_2.addWidget(self.label_tooltype)

        self.comboBox_tooltype = QComboBox(self.centralwidget)
        self.comboBox_tooltype.setObjectName(u"comboBox_tooltype")
        sizePolicy.setHeightForWidth(self.comboBox_tooltype.sizePolicy().hasHeightForWidth())
        self.comboBox_tooltype.setSizePolicy(sizePolicy)
        self.comboBox_tooltype.setMinimumSize(QSize(180, 24))
        self.comboBox_tooltype.setMaximumSize(QSize(16777215, 24))

        self.horizontalLayout_2.addWidget(self.comboBox_tooltype)

        self.label_3 = QLabel(self.centralwidget)
        self.label_3.setObjectName(u"label_3")

        self.horizontalLayout_2.addWidget(self.label_3)

        self.lineEdit_args = QLineEdit(self.centralwidget)
        self.lineEdit_args.setObjectName(u"lineEdit_args")
        sizePolicy1 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Fixed)
        sizePolicy1.setHorizontalStretch(0)
        sizePolicy1.setVerticalStretch(0)
        sizePolicy1.setHeightForWidth(self.lineEdit_args.sizePolicy().hasHeightForWidth())
        self.lineEdit_args.setSizePolicy(sizePolicy1)
        self.lineEdit_args.setMinimumSize(QSize(220, 24))
        self.lineEdit_args.setMaximumSize(QSize(5000, 24))
        self.lineEdit_args.setClearButtonEnabled(True)

        self.horizontalLayout_2.addWidget(self.lineEdit_args)


        self.verticalLayout_3.addLayout(self.horizontalLayout_2)

        self.horizontalLayout_options_placeholder = QHBoxLayout()
        self.horizontalLayout_options_placeholder.setObjectName(u"horizontalLayout_options_placeholder")

        self.verticalLayout_3.addLayout(self.horizontalLayout_options_placeholder)

        self.line_2 = QFrame(self.centralwidget)
        self.line_2.setObjectName(u"line_2")
        self.line_2.setFrameShape(QFrame.Shape.VLine)
        self.line_2.setFrameShadow(QFrame.Shadow.Sunken)

        self.verticalLayout_3.addWidget(self.line_2)

        MainWindow.setCentralWidget(self.centralwidget)
        self.statusbar = QStatusBar(MainWindow)
        self.statusbar.setObjectName(u"statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.dockWidget_program_files = QDockWidget(MainWindow)
        self.dockWidget_program_files.setObjectName(u"dockWidget_program_files")
        self.dockWidget_program_files.setAllowedAreas(Qt.BottomDockWidgetArea)
        self.dockWidgetContents_4 = QWidget()
        self.dockWidgetContents_4.setObjectName(u"dockWidgetContents_4")
        self.verticalLayout = QVBoxLayout(self.dockWidgetContents_4)
        self.verticalLayout.setSpacing(3)
        self.verticalLayout.setObjectName(u"verticalLayout")
        self.verticalLayout.setContentsMargins(3, 3, 3, 3)
        self.treeView_programfiles = SourcesTreeView(self.dockWidgetContents_4)
        self.treeView_programfiles.setObjectName(u"treeView_programfiles")
        sizePolicy2 = QSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)
        sizePolicy2.setHorizontalStretch(0)
        sizePolicy2.setVerticalStretch(100)
        sizePolicy2.setHeightForWidth(self.treeView_programfiles.sizePolicy().hasHeightForWidth())
        self.treeView_programfiles.setSizePolicy(sizePolicy2)
        font = QFont()
        font.setPointSize(11)
        self.treeView_programfiles.setFont(font)
        self.treeView_programfiles.setFocusPolicy(Qt.StrongFocus)
        self.treeView_programfiles.setAcceptDrops(True)
        self.treeView_programfiles.setLineWidth(1)
        self.treeView_programfiles.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeView_programfiles.setTextElideMode(Qt.ElideLeft)
        self.treeView_programfiles.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.treeView_programfiles.header().setVisible(False)

        self.verticalLayout.addWidget(self.treeView_programfiles)

        self.dockWidget_program_files.setWidget(self.dockWidgetContents_4)
        MainWindow.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.dockWidget_program_files)
        self.dockWidget_io_files = QDockWidget(MainWindow)
        self.dockWidget_io_files.setObjectName(u"dockWidget_io_files")
        self.dockWidget_io_files.setAllowedAreas(Qt.BottomDockWidgetArea)
        self.dockWidgetContents_2 = QWidget()
        self.dockWidgetContents_2.setObjectName(u"dockWidgetContents_2")
        self.verticalLayout_7 = QVBoxLayout(self.dockWidgetContents_2)
        self.verticalLayout_7.setSpacing(3)
        self.verticalLayout_7.setObjectName(u"verticalLayout_7")
        self.verticalLayout_7.setContentsMargins(3, 3, 3, 3)
        self.treeView_io_files = CustomTreeView(self.dockWidgetContents_2)
        self.treeView_io_files.setObjectName(u"treeView_io_files")
        self.treeView_io_files.setSelectionMode(QAbstractItemView.ExtendedSelection)
        self.treeView_io_files.setTextElideMode(Qt.ElideLeft)
        self.treeView_io_files.setVerticalScrollMode(QAbstractItemView.ScrollPerPixel)
        self.treeView_io_files.header().setVisible(False)

        self.verticalLayout_7.addWidget(self.treeView_io_files)

        self.dockWidget_io_files.setWidget(self.dockWidgetContents_2)
        MainWindow.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.dockWidget_io_files)
        self.dockWidget_program = QDockWidget(MainWindow)
        self.dockWidget_program.setObjectName(u"dockWidget_program")
        self.dockWidget_program.setFeatures(QDockWidget.DockWidgetFloatable|QDockWidget.DockWidgetMovable)
        self.dockWidget_program.setAllowedAreas(Qt.BottomDockWidgetArea)
        self.dockWidgetContents = QWidget()
        self.dockWidgetContents.setObjectName(u"dockWidgetContents")
        self.verticalLayout_8 = QVBoxLayout(self.dockWidgetContents)
        self.verticalLayout_8.setSpacing(3)
        self.verticalLayout_8.setObjectName(u"verticalLayout_8")
        self.verticalLayout_8.setContentsMargins(3, 3, 3, 3)
        self.textEdit_program = CodeTextEdit(self.dockWidgetContents)
        self.textEdit_program.setObjectName(u"textEdit_program")

        self.verticalLayout_8.addWidget(self.textEdit_program)

        self.dockWidget_program.setWidget(self.dockWidgetContents)
        MainWindow.addDockWidget(Qt.DockWidgetArea.BottomDockWidgetArea, self.dockWidget_program)

        self.retranslateUi(MainWindow)

        QMetaObject.connectSlotsByName(MainWindow)
    # setupUi

    def retranslateUi(self, MainWindow):
        MainWindow.setWindowTitle(QCoreApplication.translate("MainWindow", u"Tool Specification Editor", None))
        self.actionNew_main_program_file.setText(QCoreApplication.translate("MainWindow", u"New main program file", None))
#if QT_CONFIG(tooltip)
        self.actionNew_main_program_file.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Create new main program file</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionSelect_main_program_file.setText(QCoreApplication.translate("MainWindow", u"Select main program file", None))
#if QT_CONFIG(tooltip)
        self.actionSelect_main_program_file.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Select existing main program file from your computer</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionAdd_program_file.setText(QCoreApplication.translate("MainWindow", u"Add program files", None))
#if QT_CONFIG(tooltip)
        self.actionAdd_program_file.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Select additional program files from your computer</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionAdd_program_directory.setText(QCoreApplication.translate("MainWindow", u"Add program directory", None))
#if QT_CONFIG(tooltip)
        self.actionAdd_program_directory.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Select additional program directory from your computer</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionRemove_selected_program_files.setText(QCoreApplication.translate("MainWindow", u"Remove selected program files", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_selected_program_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Remove selected additional program files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionAdd_input_files.setText(QCoreApplication.translate("MainWindow", u"Add input files", None))
#if QT_CONFIG(tooltip)
        self.actionAdd_input_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Add input files and/or directories. Examples:</p><p>'data.csv'</p><p>'input/data.csv'</p><p>'input/'</p><p>'output/'</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionRemove_selected_input_files.setText(QCoreApplication.translate("MainWindow", u"Remove selected input files", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_selected_input_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Remove selected input files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionAdd_opt_input_files.setText(QCoreApplication.translate("MainWindow", u"Add optional input files", None))
#if QT_CONFIG(tooltip)
        self.actionAdd_opt_input_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Add optional input files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionRemove_selected_opt_input_files.setText(QCoreApplication.translate("MainWindow", u"Remove selected optional input files", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_selected_opt_input_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Remove selected optional input files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionAdd_output_files.setText(QCoreApplication.translate("MainWindow", u"Add output files", None))
#if QT_CONFIG(tooltip)
        self.actionAdd_output_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Add output files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionRemove_selected_output_files.setText(QCoreApplication.translate("MainWindow", u"Remove selected output files", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_selected_output_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Remove selected output files</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionNew_program_file.setText(QCoreApplication.translate("MainWindow", u"New program file", None))
#if QT_CONFIG(tooltip)
        self.actionNew_program_file.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Create new additional program file</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.actionSave_program_file.setText(QCoreApplication.translate("MainWindow", u"Save program file", None))
#if QT_CONFIG(tooltip)
        self.actionSave_program_file.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Save program file</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
#if QT_CONFIG(shortcut)
        self.actionSave_program_file.setShortcut(QCoreApplication.translate("MainWindow", u"Ctrl+Shift+S", None))
#endif // QT_CONFIG(shortcut)
        self.actionRemove_all_program_files.setText(QCoreApplication.translate("MainWindow", u"Remove all program files", None))
#if QT_CONFIG(tooltip)
        self.actionRemove_all_program_files.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Remove ALL program files. Useful with Executable Tool Specs.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.label_tooltype.setText(QCoreApplication.translate("MainWindow", u"Tool type:", None))
#if QT_CONFIG(tooltip)
        self.comboBox_tooltype.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Tool specification type</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.comboBox_tooltype.setCurrentText("")
        self.label_3.setText(QCoreApplication.translate("MainWindow", u"Command line arguments:", None))
#if QT_CONFIG(tooltip)
        self.lineEdit_args.setToolTip(QCoreApplication.translate("MainWindow", u"<html><head/><body><p>Command line arguments (space-delimited) for the main program (optional). Use '@@' tags to refer to input files or URLs, see the User Guide for details.</p></body></html>", None))
#endif // QT_CONFIG(tooltip)
        self.lineEdit_args.setPlaceholderText(QCoreApplication.translate("MainWindow", u"Type arguments here...", None))
        self.dockWidget_program_files.setWindowTitle(QCoreApplication.translate("MainWindow", u"Program files", None))
        self.dockWidget_io_files.setWindowTitle(QCoreApplication.translate("MainWindow", u"Input && output files", None))
        self.dockWidget_program.setWindowTitle("")
    # retranslateUi




######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

""" Spine items. """
from .version import __version__


def _factories_and_executable_items():
    from . import data_connection, data_store, data_transformer, exporter, importer, merger, tool, view
    from .data_connection.data_connection_factory import DataConnectionFactory
    from .data_store.data_store_factory import DataStoreFactory
    from .data_transformer import specification_factory
    from .data_transformer.data_transformer_factory import DataTransformerFactory
    from .exporter import specification_factory
    from .exporter.exporter_factory import ExporterFactory
    from .importer import specification_factory
    from .importer.importer_factory import ImporterFactory
    from .merger.merger_factory import MergerFactory
    from .tool import specification_factory
    from .tool.tool_factory import ToolFactory
    from .view.view_factory import ViewFactory

    modules = (data_connection, data_store, data_transformer, exporter, importer, merger, tool, view)
    item_infos = tuple(module.item_info.ItemInfo for module in modules)
    factories = (
        DataConnectionFactory,
        DataStoreFactory,
        DataTransformerFactory,
        ExporterFactory,
        ImporterFactory,
        MergerFactory,
        ToolFactory,
        ViewFactory,
    )
    factories = {info.item_type(): factory for info, factory in zip(item_infos, factories)}
    executables = {module.item_info.ItemInfo.item_type(): module.executable_item.ExecutableItem for module in modules}
    specification_item_submodules = (data_transformer, exporter, importer, tool)
    specification_factories = {
        module.item_info.ItemInfo.item_type(): module.specification_factory.SpecificationFactory
        for module in specification_item_submodules
    }
    return (
        factories.copy,
        executables.copy,
        specification_factories.copy,
    )


item_factories, executable_items, item_specification_factories = _factories_and_executable_items()
del _factories_and_executable_items



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Contains a table model that can be used as drop target."""
import pickle
from PySide6.QtCore import QAbstractTableModel
from ..commands import InsertRow
from ..widgets.drop_target_table import DROP_MIME_TYPE


class ParameterDropTargetTableModel(QAbstractTableModel):
    def dropMimeData(self, data, action, row, column, parent):
        if row < 0:
            row = self.rowCount()
        rows = []
        parameters = pickle.loads(data.data(DROP_MIME_TYPE))
        for entity_class, parameter_list in parameters.items():
            for parameter in parameter_list:
                rows.append(self._make_drop_row(entity_class, parameter))
        if len(rows) == 1:
            self._undo_stack.push(InsertRow("add parameter", self, row, rows[0]))
        else:
            self._undo_stack.beginMacro("add parameters")
            for i, row_data in enumerate(rows):
                self._undo_stack.push(InsertRow("", self, row + i, row_data))
            self._undo_stack.endMacro()
        return True

    def mimeTypes(self):
        return [DROP_MIME_TYPE]

    @staticmethod
    def _make_drop_row(entity_class, parameter):
        raise NotImplementedError()



######################################################################################################################
# Copyright (C) 2017-2022 Spine project consortium
# Copyright Spine Items contributors
# This file is part of Spine Items.
# Spine Items is free software: you can redistribute it and/or modify it under the terms of the GNU Lesser General
# Public License as published by the Free Software Foundation, either version 3 of the License, or (at your option)
# any later version. This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
# without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General
# Public License for more details. You should have received a copy of the GNU Lesser General Public License along with
# this program. If not, see <http://www.gnu.org/licenses/>.
######################################################################################################################

"""Widget shown to user when a new Tool is created."""
from spinetoolbox.widgets.add_project_item_widget import AddProjectItemWidget
from ..item_info import ItemInfo
from ..tool import Tool


class AddToolWidget(AddProjectItemWidget):
    """A widget that queries user's preferences for a new item.

    Attributes:
        toolbox (ToolboxUI): Parent widget
        x (int): X coordinate of new item
        y (int): Y coordinate of new item
        spec (str): Tool specification
    """

    def __init__(self, toolbox, x, y, spec=""):
        """Initialize class."""
        super().__init__(toolbox, x, y, Tool, spec=spec)

    def call_add_item(self):
        """Creates new Item according to user's selections."""
        spec = self.ui.comboBox_specification.currentText()
        item = {
            self.name: {
                "type": ItemInfo.item_type(),
                "description": self.description,
                "x": self._x,
                "y": self._y,
                "specification": spec,
                "execute_in_work": True,
            }
        }
        self._toolbox.add_project_items(item)
